#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  Commands and their actions.
"""

import asyncio
import logging

class Commands:
    """ Commands for IRC.

    Each method in this class is a command to run. It should complete the 
    action required by the command.
    """
    log = logging.getLogger('ad1459.commands')
    def me(self, room, network, message):
        """ /me command

        Sends a CTCP Action.

        Arguments:
            room (:obj:`Room`): The room to ACTION
            client (:obj:`Client`): The client object
            message (str): The action to send.
        """
        self.log.debug('Sending action to %s: %s', room.name, message)
        asyncio.run_coroutine_threadsafe(
            network.client.ctcp(room.name, 'ACTION', contents=message),
            loop=asyncio.get_event_loop()
        )
    
    def join(self, room, network, command):
        """ /join command

        Joins a room on IRC.

        Arguments: 
            room : unused.
            client (:obj:'Client`): The client we're using to join.
            command (str): The text of the command.
        """
        channels = command.split()

        for channel in channels:
            if channel.startswith('#'):
                network.join_channel(channel)
    
    def quit(self, room, network, command):
        """ /quit command

        Disconnects from IRC.
        """
        if command:
            network.disconnect(message=command)
        else:
            network.disconnect()

    def part(self, room, network, command):
        if command:
            asyncio.run_coroutine_threadsafe(
                network.client.part(room.name, message=command),
                loop=asyncio.get_event_loop()
            )
        else:
            asyncio.run_coroutine_threadsafe(
                network.client.part(room.name),
                loop=asyncio.get_event_loop()
            )
    
    def kick(self, room, network, command):
        com = command.split()
        user = com.pop(0)
        reason = ' '.join(com)
        if network.nickname in room.ops:
            if reason:
                asyncio.run_coroutine_threadsafe(
                    network.client.kick(room.name, user, reason=reason),
                    loop=asyncio.get_event_loop()
                )
            else:
                asyncio.run_coroutine_threadsafe(
                    network.client.kick(room.name, user),
                    loop=asyncio.get_event_loop()
                )
        else:
            room.add_message(
                f'You are not an op in {room.name}', kind='notice'
            )
    
    def ban(self, room, network, command):
        com = command.split()
        user = com.pop(0)
        reason = ' '.join(com)
        if network.nickname in room.ops:
            asyncio.run_coroutine_threadsafe(
                network.client.ban(room.name, user),
                loop=asyncio.get_event_loop()
            )
        else:
            room.add_message(
                f'You are not an op in {room.name}', kind='notice'
            )
    
    def kickban(self, room, network, command):
        com = command.split()
        user = com.pop(0)
        reason = ' '.join(com)
        if network.nickname in room.ops:
            if reason:
                asyncio.run_coroutine_threadsafe(
                    network.client.kickban(room.name, user, reason=reason),
                    loop=asyncio.get_event_loop()
                )
            else:
                asyncio.run_coroutine_threadsafe(
                    network.client.kickban(room.name, user),
                    loop=asyncio.get_event_loop()
                )
        else:
            room.add_message(
                f'You are not an op in {room.name}', kind='notice'
            )
    
    def nick(self, room, network, command):
        network.change_nick(command.split()[0])
    
    def query(self, room, network, command):
        com = command.split()
        user = com.pop(0)
        message = ' '.join(com)
        msgroom = network.get_room_for_name(user)
        network.window.switcher.switcher.select_row(msgroom.row)
        if message:
            network.send_message(msgroom, message)
    
    def msg(self, room, network, command):
        com = command.split()
        user = com.pop(0)
        message = ' '.join(com)
        msgroom = network.get_room_for_name(user)
        if message:
            network.send_message(msgroom, message)
    
    def notice(self, room, network, command):
        com = command.split()
        target = com.pop(0)
        notice = ' '.join(com)
        msgroom = network.get_room_for_name(target)
        if notice:
            asyncio.run_coroutine_threadsafe(
                network.client.notice(target, notice),
                loop=asyncio.get_event_loop()
            )
    
    def topic(self, room, network, command):
        com = command.split()
        if com[0].startswith('#'):
            channel = com.pop(0)
            topic = ' '. join(com)
        else:
            channel = room.name
            topic = command
        
        asyncio.run_coroutine_threadsafe(
            network.client.set_topic(channel, topic.strip()),
            loop=asyncio.get_event_loop()
        )



#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  This is the About Dialog.
"""

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from ad1459 import __version__, __license__


class AboutDialog(Gtk.AboutDialog):

    def __init__(self):
        super().__init__()

        self.set_program_name('AD1459')
        self.set_logo_icon_name('in.donotspellitgav.Ad1459')
        self.set_version(__version__.__version__)
        self.set_website('https://github.com/g4vr0che/ad1459')
        self.set_website_label('github.com/g4vr0che/ad1459')
        self.set_copyright('Copyright ©2019-2020 Gaven Royer et al')
        self.set_license(__license__.__license__)
        self.set_wrap_license(True)
        self.set_authors(
            [
                'Gaven Royer'
            ]
        )



#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  This file is an application window.
"""

import logging

import gi
gi.require_versions(
    {
        'Gtk': '3.0',
        'Gdk': '3.0'
    }
)
from gi.repository import Gtk, Gdk

from .. import formatting
from .about import AboutDialog
from .headerbar import Headerbar
from .irc_entry import IrcEntry
from .room_switcher import RoomSwitcher
from .server_popup import ServerPopover

class Ad1459Window(Gtk.ApplicationWindow):
    """ This is a window for AD1459. 
    
    This contains the overall layout as well as most of the general controls for
    the app.

    Attributes:
        active_room (:obj:`Room`): The currently active room.
        focused (bool): Whether the window is currently focused or not.
    """

    def __init__(self, app):
        self.log = logging.getLogger('ad1459.window')
        self.log.debug('Creating window')

        self.app = app
        self.parser = formatting.Parser()

        super().__init__()
        self.set_default_size(1000, 600)

        self.header = Headerbar(self.app)
        self.set_titlebar(self.header)

        self.add_help_overlay()

        self.main_pane = Gtk.HPaned()
        self.main_pane.set_position(200)
        self.add(self.main_pane)

        channel_grid = Gtk.Grid()
        channel_grid.set_hexpand(True)
        channel_grid.set_vexpand(True)
        self.main_pane.add2(channel_grid)

        switcher_grid = Gtk.Grid()
        switcher_grid.set_hexpand(True)
        switcher_grid.set_vexpand(True)
        self.main_pane.add1(switcher_grid)

        self.switcher = RoomSwitcher(self.app)
        switcher_grid.attach(self.switcher, 0, 0, 1, 1)

        join_bar = Gtk.ActionBar()
        switcher_grid.attach(join_bar, 0, 1, 1, 1)

        self.join_entry = Gtk.Entry()
        self.join_entry.set_width_chars(1)
        self.join_entry.set_hexpand(True)
        self.join_entry.set_placeholder_text('Join channel')
        self.join_entry.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, 'list-add-symbolic'
        )
        self.join_entry.set_icon_activatable(
            Gtk.EntryIconPosition.SECONDARY, True
        )
        join_bar.pack_start(self.join_entry)

        entry_grid = Gtk.ActionBar()
        channel_grid.attach(entry_grid, 0, 1, 1, 1)
        
        self.irc_entry = IrcEntry(self)
        self.nick_button = Gtk.Button.new_with_label('nickname')
        self.send_button = Gtk.Button.new_from_icon_name(
            'mail-send-symbolic', Gtk.IconSize.BUTTON
        )

        self.nick_button.set_halign(Gtk.Align.START)
        self.send_button.set_halign(Gtk.Align.END)

        Gtk.StyleContext.add_class(
            self.nick_button.get_style_context(), 'flat'
        )
        Gtk.StyleContext.add_class(
            self.send_button.get_style_context(), 'suggested-action'
        )

        entry_grid.pack_start(self.nick_button)
        entry_grid.pack_start(self.irc_entry)
        entry_grid.pack_start(self.send_button)

        self.channel_pane = Gtk.HPaned()
        self.channel_pane.set_position(600)
        channel_grid.attach(self.channel_pane, 0, 0, 1, 1)

        self.message_stack = Gtk.Stack()
        self.message_stack.set_hexpand(True)
        self.message_stack.set_vexpand(True)
        self.message_stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.message_stack.set_transition_duration(100)
        self.channel_pane.pack1(self.message_stack, True, True)

        self.topic_stack = Gtk.Stack()
        self.topic_stack.set_hexpand(True)
        self.topic_stack.set_vexpand(True)
        self.topic_stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.topic_stack.set_transition_duration(100)
        self.channel_pane.pack2(self.topic_stack, False, False)

        self.irc_entry.grab_focus_without_selecting()

        self.recents = []

        self.connect('focus-in-event', self.on_focus_changed)
    
    # Methods
    def add_help_overlay(self):
        keys_window = Gtk.ShortcutsWindow()
        main_section = Gtk.ShortcutsSection()
        main_section.props.section_name = 'AD1459 Shortcuts'
        keys_window.add(main_section)
        
        entry_group = Gtk.ShortcutsGroup()
        entry_group.props.title = 'Entry Shortcuts'
        main_section.add(entry_group)
        
        tab_shortcut = Gtk.ShortcutsShortcut()
        tab_shortcut.props.accelerator = 'Tab'
        tab_shortcut.props.title = (
            'Auto-complete nicknames'
        )
        entry_group.add(tab_shortcut)
        
        room_recent_shortcut = Gtk.ShortcutsShortcut()
        room_recent_shortcut.props.accelerator = 'Up Down'
        room_recent_shortcut.props.title = (
            'Move through the recently sent messages for the current channel'
        )
        entry_group.add(room_recent_shortcut)
        
        win_recent_shortcut = Gtk.ShortcutsShortcut()
        win_recent_shortcut.props.accelerator= '<ctl>Up <clt>Down'
        win_recent_shortcut.props.title = (
            'Move through the recently sent messages from all channels'
        )
        entry_group.add(win_recent_shortcut)

        self.set_help_overlay(keys_window)
    
    # Internal Handlers
    def on_focus_changed(self, window, data=None):
        """ Unset the unread indicator for the current room when we gain focus.
        """
        try:
            room = self.message_stack.get_visible_child().room
        except AttributeError:
            return
        self.log.debug('Unsetting unread indicator for %s', room.name)
        room.row.icon = 'none'

    # Data
    @property
    def active_room(self):
        """ Room: The currently active room."""
        room = self.message_stack.get_visible_child().room
        return room

    @property
    def focused(self):
        """bool: Whether or not the window has focus."""
        if self.props.is_active:
            return True
        
        return False



#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  The topic pane contents.
"""

import logging

import gi
gi.require_versions(
    {
        'Gtk': '3.0'
    }
)
from gi.repository import Gtk, GLib

from .user_row import UserRow
from .room_row import RoomKind

class TopicPane(Gtk.Grid):
    """ The right-sidebar of the UI, containing the topic and the User List."""

    def __init__(self, room):
        super().__init__()
        self.room = room
        self.log = logging.getLogger('ad1459.topic')
        self.topic_expander = Gtk.Expander()
        self.topic_expander.set_margin_top(6)
        self.topic_expander.set_margin_end(3)
        self.topic_expander.set_label_fill(True)
        self.topic_expander.connect('notify::expanded', self.show_hide_topic)
        self.attach(self.topic_expander, 0, 0, 1, 1)

        # Set Expander Label
        self.exp_label = Gtk.Label()
        self.exp_label.set_line_wrap(True)
        self.exp_label.set_xalign(0)
        self.exp_label.set_margin_start(3)
        self.exp_label.set_margin_end(3)
        self.topic_expander.set_label_widget(self.exp_label)

        self.revealer = Gtk.Revealer()
        self.revealer.set_margin_start(18)
        self.revealer.set_margin_end(3)
        self.revealer.set_transition_type(Gtk.RevealerTransitionType.SLIDE_DOWN)
        self.attach(self.revealer, 0, 1, 1, 1)

        self.topic_label = Gtk.Label()
        self.topic_label.set_margin_start(6)
        self.topic_label.set_margin_end(6)
        self.topic_label.set_margin_top(3)
        self.topic_label.set_margin_bottom(6)
        self.topic_label.set_xalign(0)
        self.topic_label.set_line_wrap(True)
        self.revealer.add(self.topic_label)

        separator = Gtk.Separator.new(Gtk.Orientation.HORIZONTAL)
        separator.set_hexpand(True)
        self.attach(separator, 0, 2, 1, 1)

        user_window = Gtk.ScrolledWindow()
        user_window.set_hexpand(True)
        user_window.set_vexpand(True)
        self.attach(user_window, 0, 3, 1, 1)

        self.user_list = Gtk.ListBox()
        self.user_list.set_hexpand(True)
        self.user_list.set_vexpand(True)
        self.user_list.set_activate_on_single_click(False)
        self.user_list.set_sort_func(sort_users)
        self.user_list.connect(
            'row-activated',
            self.on_user_row_activated,
            self.room.window
        )
        user_window.add(self.user_list)

        self.update_topic()
    
    def update_topic(self):
        """Update the channel topic."""
        self.log.debug('Updating topic for %s', self.room.name)

        try:
            topic_expander_label = '<small>Set by '
            topic_by = GLib.markup_escape_text(self.room.data['topic_by'])
            topic_expander_label += f'<i>{topic_by}</i> on '
            topic_expander_label += f'<i>{self.room.data["topic_set"]}</i>:</small>'
            topic_text = GLib.markup_escape_text(self.room.data['topic'])
            topic_text = self.room.window.parser.parse_text(topic_text)
            topic_text = self.room.window.parser.hyperlinks(topic_text)
            self.topic_expander.set_sensitive(True)
            self.topic_expander.set_expanded(True)
        
        except (KeyError, TypeError):
            topic_expander_label = "No topic set"
            topic_text = ''
            self.topic_expander.set_sensitive(False)
            self.topic_expander.set_expanded(False)

        self.exp_label.set_markup(topic_expander_label)
        self.topic_label.set_markup(f'<small>{topic_text}</small>')

    def update_users(self):
        """Updates this room's user list."""
        curr_users = self.user_list.get_children()
        
        for user in curr_users:
            GLib.idle_add(user.destroy)

        for user in self.room.users:
            new_user = UserRow(self.room)
            new_user.nick = user
            self.user_list.add(new_user)
            modes = []
            
            if user in self.room.mutes:
                modes.append('q')
            
            if user in self.room.voices:
                modes.append('v')

            if user in self.room.ops:
                modes.append('o')
            
            new_user.modes = modes
            
            if user == 'Miga':
                self.log.debug('Miga mdoes: %s', new_user.modes)
        
        self.room.window.show_all()
        self.user_list.invalidate_sort()
    
    # Internal Handlers
    def show_hide_topic(self, expander, data=None):
        """ Shows or hides the topic.

        This is synced to the `expanded` property of `expander`.
        """
        self.revealer.props.reveal_child = expander.get_expanded()
    
    def on_user_row_activated(self, list_box, row, window):
        """ `row-activated` handler for the list bow. 

        This opens a PM room with the selected user.
        """
        user = row.nick
        network = row.room.network
        room = network.get_room_for_name(user)
        window.message_stack.set_visible_child_name(room.id)
        window.topic_stack.set_visible_child_name(room.id)
        window.irc_entry.grab_focus_without_selecting()
        window.nick_button.set_label(network.nickname)
        window.switcher.switcher.select_row(room.row)

def sort_users(row1, row2, *user_data):
    """ Sorts the users for the user list.

    Arguments:
        row1: The first user row to compare
        row2: The second user row to compare
    
    Returns:
        -1 if row1 should go above row2
        1 if row1 should go below row2
    """
    if row1.mute:
        if row2.mute:
            return sort_users_alpha(row1, row2)
        
        else:
            return 1
    
    if row1.voice:
        if row2.voice:
            return sort_users_alpha(row1, row2)
        
        elif row2.op:
            return 1
        
        else:
            return -1
    
    if row1.op:
        if row2.op:
            return sort_users_alpha(row1, row2)
        
        else:
            return -1
    
    if row1.nomodes:
        if row2.nomodes:
            return sort_users_alpha(row1, row2)
        
        elif row2.mute:
            return -1
        
        else:
            return 1

def sort_users_alpha(row1, row2):
    """ Sorts the users alphabetically by nick."""

    if row1.nick.upper() < row2.nick.upper():
        return -1
    else:
        return 1


#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  The IRC client object.
"""

import asyncio
import logging
import pydle
import time

class Client(pydle.Client):
    """ This is the actual IRC Client. 

    Most of this functionality is provided by Pydle.
    """

    def __init__(self, nick, network, sasl_username=None, sasl_password=None, **kwargs):
        self.log = logging.getLogger('ad1459.client')
        super().__init__(nick, sasl_username=sasl_username, sasl_password=sasl_password, **kwargs)
        self.network_ = network
        self.log.debug('Created client for network %s', self.network_.name)
        self.bouncer = False
    
    async def connect(self, hostname=None, password=None, **kwargs):
        self.log.debug('Client initiating connection to %s', hostname)
        await super().connect(hostname=hostname, password=password, **kwargs)
    
    async def on_connect(self):
        self.log.info('Connected to %s', self.network_.name)
        await super().on_connect()
        await self.network_.on_connected()
    
    async def _disconnect(self, expected):
        """ We need to override this to stop the event loop getting closed."""
        # Shutdown connection.
        await self.connection.disconnect()

        # Reset any attributes.
        self._reset_attributes()

        # Callback.
        await self.on_disconnect(expected)

        """This is our override.
        # Shut down event loop.
        # if expected and self.own_eventloop:
        #     self.connection.stop()
        """
    
    async def on_raw(self, message):
        if message.command == ('CAP' or 'cap' or 'Cap'):
            if 'znc.in/' in " ".join(message.params):
                self.log.debug('Server appears to be a ZNC Bouncer')
                self.bouncer = True
        
        await super().on_raw(message)
    
    async def on_nick_change(self, old, new):
        self.log.debug('User %s is now %s', old, new)
        await self.network_.on_nick_change(old, new)
        await super().on_nick_change(old, new)
    
    async def on_join(self, channel, user):
        self.log.debug(f'User {user} joined {channel} on {self.network_.name}')
        await self.network_.on_join(channel, user)
        await super().on_join(channel, user)
    
    async def on_part(self, channel, user, message=None):
        self.log.debug(f'User {user} parted {channel} on {self.network_.name}')
        await self.network_.on_part(channel, user, message=message)
        await super().on_part(channel, user, message=message)
    
    async def on_kick(self, channel, target, by, reason=None):
        self.log.debug('%s was kicked from %s by %s (%s)', target, channel, by, reason)
        await self.network_.on_kick(channel, target, by, reason=reason)
        await super().on_kick(channel, target, by, reason=reason)
    
    async def on_quit(self, user, message=None):
        self.log.debug(f'User {user} has quit {self.network_.name}')
        await self.network_.on_quit(user, message=message)
        await super().on_quit(user, message=message)
    
    async def on_kill(self, target, by, reason):
        self.log.debug('%s was killed by %s, (%s)', target, by, reason)
        await self.network_.on_kill(target, by, reason=reason)
        super().on_kill(target, by, reason)

    async def on_channel_message(self, target, source, message):
        self.log.debug('New message in %s from %s: %s', target, source, message)
        await self.network_.on_message(target, source, message)
        await super().on_message(target, source, message)
    
    async def on_channel_notice(self, target, by, message):
        self.log.debug('Received notice to %s from %s', target, by)
        await self.network_.on_notice(target, by, message)
        await super().on_notice(target, by, message)
    
    async def on_private_message(self, target, by, message):
        self.log.debug('New private message to %s from %s: %s', target, by, message)
        await self.network_.on_private_message(target, by, message)
        await super().on_private_message(target, by, message)
    
    async def on_private_notice(self, target, by, message):
        self.log.debug('Received private notice to %s from %s', target, by)
        await self.network_.on_private_notice(target, by, message)
        await super().on_private_notice(target, by, message)
    
    async def on_mode_change(self, channel, modes, by):
        self.log.debug('%s set modes %s in %s', by, modes, channel)
        await self.network_.on_mode_change(channel, modes, by)
        await super().on_mode_change(channel, modes, by)
    
    async def on_topic_change(self, channel, message, by):
        self.log.debug('Topic in %s set to "%s" by %s', channel, message, by)
        await self.network_.on_topic_change(channel, message, by)
        await super().on_topic_change(channel, message, by)
    
    async def on_user_invite(self, target, channel, by):
        self.log.debug('%s was invited to %s by %s', target, channel, by)
        await self.network_.on_user_invite(target, channel, by)
        super().on_user_invite(target, channel, by)

    async def on_invite(self, channel, by):
        self.log.debug('Invited to %s by %s', channel, by)
        await self.network_.on_invite(channel, by)
        await super().on_invite(channel, by)
    
    # CTCP Support

    async def on_ctcp(self, by, target, what, contents):
        self.log.debug('%s received CTCP %s from %s: %s', target, what, by, contents)
        await super().on_ctcp(by, target, what, contents)
    
    async def on_ctcp_reply(self, by, target, what, contents):
        self.log.debug('%s received REPLY CTCP %s from %s: %s', target, what, by, contents)
        await super().on_ctcp(by, target, what, contents)

    async def on_ctcp_action(self, by, target, contents):
        await self.network_.on_ctcp_action(target, by, contents)


#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  Handling for rooms.
"""

import asyncio
import enum
import logging
import time as Time

import gi
gi.require_version('Notify', '0.7')
from gi.repository import Notify

from .widgets.message_buffer import MessageBuffer
from .widgets.message_row import MessageRow
from .widgets.room_row import RoomRow, RoomKind, RoomIcon
from .widgets.topic import TopicPane

class Room:
    """ A Room object that represents a list of grouped messages on IRC. This 
    can be a channel, PM conversation/dialog/query, or a list of 
    server messages.
    
    Attributes:
        buffer (:obj:): The message buffer for this room.
        topic (:obj:): The room topic (if applicable).
        row (:obj:`RoomRow`) The row for this room in the switcher.
    """

    def __init__(self, app, network, window, name):
        Notify.init('AD1459')
        self.log = logging.getLogger('ad1459.room')
        self.log.debug('Creating new room %s', name)

        self.app = app
        self.network = network 
        self.window = window

        self.name = name

        self.notification = Notify.Notification.new('AD1459', 'Init')
        self.buffer = MessageBuffer(self)
        self.row = RoomRow(self)
        self.topic_pane = TopicPane(self)
        self.old_users = []
        self._tab_complete = []
        self.unsent_text = ''
        self.recents = []
    
    # Methods
    def add_message(self, message, sender='*', kind='message', time=None):
        """ Adds a message into this room and inserts it into the buffer.

        Arguments:
            message (str): the text of the message to add.
            sender (str): The person/entity who sent the message
            time (str): The time the message was sent.
            kind (str): The type of message this is (default: 'message')
        """
        if not time:
            time = Time.ctime().split()[3]
        
        if (self.network.nickname in message and kind != 'server'):
            kind = 'highlight'
        
        if sender == self.network.nickname and kind != 'server':
            kind = 'mine'

        # Create the new message
        new_message = MessageRow()
        new_message.time = time
        new_message.sender = sender
        new_message.text = message
        new_message.kind = kind

        # Get the same of the unread Icon
        current_room = self.window.message_stack.get_visible_child().room
        
        # This sets the unread indicator for the channel
        if self.window.active_room != self or not self.window.focused:
            if self.row.icon.value <= RoomIcon[kind.upper()].value:
                self.row.icon = kind
        
        # TODO: This can be improved
        # This shows the notification
        if not self.window.focused:
            if (
                    new_message.kind == 'highlight' or 
                    (self.kind == RoomKind.DIALOG and
                    new_message.kind != 'mine')
            ):
                if self.kind == RoomKind.CHANNEL:
                    self.notification.update(
                        f'{sender} mentioned you in {self.name}',
                        message
                    )
                
                else:
                    self.notification.update(
                        f'New message from {sender}',
                        message
                    )
                self.notification.show()
        
        # This posts the message, in either the last server message or a new one
        messages = self.buffer.list_box.get_children()
        if messages:
            last_message = messages[-1]
            if last_message:
                if kind == 'server' and last_message.kind == 'server':
                    last_message.update_server_message(message)
                    last_message.time = time
                
                else:
                    last_message.server_message_expander.set_expanded(False)
                    self.buffer.add_message_to_buffer(new_message)

            else:
                last_message.server_message_expander.set_expanded(False)
                self.buffer.add_message_to_buffer(new_message)
        
        else:
            self.buffer.add_message_to_buffer(new_message)
        
        self.window.show_all()
    
    def update_tab_complete(self, user):
        """Add/move a user to the end of the tab-complete list

        Arguments:
            user (str): the nick of the user to move.
        """
        if user in self._tab_complete:
            self._tab_complete.remove(user)
        
        self._tab_complete.append(user)
    
    def update_users(self):
        """ Updates the user list in the UI."""
        self.old_users = []
        for user in self.users:
            self.old_users.append(user)
            
        self.topic_pane.update_users()
    
    def leave(self):
        """ Remove this room from the UI."""
        self.log.debug('Removing room %s form list', self.name)
        self.buffer.destroy()
        self.topic_pane.destroy()
        self.row.destroy()
        self.notification.close()
    
    def part(self):
        """ Sends a part from this channel.
        
        Note that we don't remove the room until we receive the PART message
        from the server.
        """
        asyncio.run_coroutine_threadsafe(
            self.network.client.part(self.name, message='Leaving...'),
            loop=asyncio.get_event_loop()
        )

    # Data
    @property
    def data(self):
        """dict: A dictionary with this channel's data."""
        try:
            return self.network.client.channels[self.name]
        except KeyError:
            return {
                'users': {self.name, self.network.nickname},
                'topic': f'Private chat with {self.name}',
                'topic_by': self.network.name,
                'topic_set': Time.ctime().split()[3]
            }
    
    @property
    def users(self):
        """list: A list of users in this room."""
        users = []

        if self.kind == RoomKind.SERVER:
            users.append(self.network.nickname)
        
        elif self.kind == RoomKind.DIALOG:
            users.append(self.name)
        
        else:
            for user in self.data['users']:
                users.append(user)

        return users
    
    @property
    def ops(self):
        """:list: of str: The list of chanops in this room."""
        if self.kind == RoomKind.CHANNEL:
            try:
                return self.data['modes']['o']
            
            except KeyError:
                return []
        else:
            return []

    
    @property
    def voices(self):
        if self.kind == RoomKind.CHANNEL:
            try:
                return self.data['modes']['v']

            except KeyError:
                return []
        else:
            return []

    
    @property
    def mutes(self):
        if self.kind == RoomKind.CHANNEL:
            try: 
                return self.data['modes']['q']

            except KeyError:
                return []
        else:
            return []

    
    @property
    def tab_complete(self):
        return self._tab_complete

    @property
    def kind(self):
        """:obj:`RoomKind` The type of room this is."""
        return self._kind
    
    @kind.setter
    def kind(self, kind):
        self.log.debug('Setting room type for %s to %s', self.name, kind)
        kind = kind.upper()
        self._kind = RoomKind[kind]
        self.row.set_margins()
        self.log.debug('Set margins for %s room', kind)

    @property
    def name(self):
        """str: The name of this channel. This is displayed in the UI, and is
        the name of the channel, the network, or the user we are chatting with.
        """
        return self._name
    
    @name.setter
    def name(self, name):
        self._name = name
    
    @property
    def id(self):
        """str: The internal ID of this channel. This is used internally to 
        identify this room and it's children within the UI.

        We use the id() function to get a unique identifier for self, since this
        is guaranteed to be unique while the object exists.
        """
        return f'{self.name}-{id(self)}'



#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  ListBoxRows for networks/rooms.
"""

from enum import Enum

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

class RoomKind(Enum):
    """ An enum to classify the type of room this is."""
    SERVER = 1
    CHANNEL = 2
    ROOM = 2
    PRIVMSG = 3
    QUERY = 3
    DIALOG = 3
    WHISPER = 3

    def __str__(self):
        """ Turn this back into a string. """
        strings = {
            1: 'network',
            2: 'channel',
            3: 'user'
        }
        return strings[self.value]

class RoomIcon(Enum):
    """ Classify the icon each RoomRow uses."""
    NONE = 1
    SERVER = 2
    MESSAGE = 3
    ACTION = 3
    NOTICE = 3
    MINE = 3
    HIGHLIGHT = 4

    def __str__(self):
        strings = {
            1: 'radio-symbolic',
            2: 'radio-mixed-symbolic',
            3: 'radio-checked-symbolic',
            4: 'emblem-important-symbolic'
        }
        return strings[self.value]

def sort_by_server(row1, row2):
    """Sorts the rows alphabetically by their network name."""
    if row1.room.network.name < row2.room.network.name:
        return -1
    
    else:
        return 1

def sort_by_room(row1, row2):
    """ Sorts the rows alphabetically by room name."""
    if row1.room.name < row2.room.name:
        return -1
    
    else:
        return 1

def room_row_sort(row1, row2, *user_data):
    """ Tells whether row1 should be before row2 or not

    Arguments:
        row1 (:obj:`RoomRow`): The first row to compare
        row2 (:obj:`RoomRow`): The second row to compare
        user_data (:obj:`object or None): user data
    
    Returns:
        -1 if row1 should go before row2.
        0  if they are equal and the order doesn't matter. 
        1  if row1 should go after row2.
    """
    
    if row1.kind == RoomKind.SERVER:
        if row2.kind == RoomKind.SERVER:
            return sort_by_server(row1, row2)
        else:
            
            if row2.room in row1.room.network.rooms:
                return -1

            else:
                return sort_by_server(row1, row2)

    elif row1.kind == RoomKind.CHANNEL:
        if row2.kind == RoomKind.SERVER:
            if row1.room in row2.room.network.rooms:
                return 1
            
            else:
                return sort_by_server(row1, row2)

        elif row2.kind == RoomKind.CHANNEL:
            if row1.room in row2.room.network.rooms:
                return sort_by_room(row1, row2)
            
            else:
                return sort_by_server(row1, row2)

        elif row2.kind == RoomKind.DIALOG:
            if row1.room in row2.room.network.rooms:
                return -1
            
            else:
                return sort_by_server(row1, row2)

    elif row1.kind == RoomKind.DIALOG:
        if row2.kind == RoomKind.SERVER or row2.kind == RoomKind.CHANNEL:
            if row1.room in row2.room.network.rooms:
                return 1

            else:
                
                return sort_by_server(row1, row2)

        if row2.kind == RoomKind.DIALOG:
            if row1.room in row2.room.network.rooms:
                return sort_by_room(row1, row2)
            
            else:
                
                return sort_by_server(row1, row2)

class RoomRow(Gtk.ListBoxRow):
    """ A dedicated ListBoxRow for representing a room in the Room List.
    
    Attributes:
        room_name (str): The name of this room.
        kind (RoomKind Enum): The type of room this is.
    """

    def __init__(self, room):
        super().__init__()
        self.room = room
        self._icon = RoomIcon.NONE
        self.set_can_focus(False)

        grid = Gtk.Grid()
        grid.set_column_spacing(6)
        grid.set_margin_top(3)
        grid.set_margin_bottom(3)
        grid.set_margin_start(3)
        grid.set_margin_end(3)
        self.add(grid)

        self.room_label = Gtk.Label()
        self.room_label.set_text(self.room.name)
        grid.attach(self.room_label, 0, 0, 1, 1)

        self.unread_indicator = Gtk.Image.new_from_icon_name(
            'radio-symbolic',
            Gtk.IconSize.SMALL_TOOLBAR            
        )
        grid.attach(self.unread_indicator, 1, 0, 1, 1)

    # Methods
    def set_margins(self):
        """Sets the margins on self depending on what type of room we are."""
        print(self.room.kind)
        if self.room.kind == RoomKind.SERVER:
            self.set_margin_top(12) 
            self.set_margin_start(0)
        elif self.room.kind == RoomKind.CHANNEL: 
            self.set_margin_top(0)
            self.set_margin_start(12)
        elif self.room.kind == RoomKind.DIALOG:
            self.set_margin_top(0)
            self.set_margin_start(18)
        else:
            self.set_margin_top(0)
            self.set_margin_start(0)

    def set_icon(self, icon_name):
        self.unread_indicator.set_from_icon_name(
            icon_name, Gtk.IconSize.SMALL_TOOLBAR
        )

    # Data
    @property
    def kind(self):
        """Easier compatibility with sort func."""
        return self.room.kind
    
    @property
    def icon(self):
        """`RoomIcon`: The icon that this row should have."""
        return self._icon
    
    @icon.setter
    def icon(self, icon):
        try:
            _icon = str(icon).upper()
            self._icon = RoomIcon[_icon]
        
        except AttributeError:
            self._icon = RoomIcon(icon)
        
        self.unread_indicator.set_from_icon_name(
            str(self._icon),
            Gtk.IconSize.SMALL_TOOLBAR
        )
   


#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  A parser for IRC style tags.
"""

import logging
from urllib.parse import urlparse

class Parser:
    """ This is a text formatting helper class.

    It aids with IRC formatting as well as providing Pango markup to the 
    interface for label widgets.
    """

    formatting = {
        '\x02': ['b', '*'],
        #'\u0003': 'color',
        #'\u000F': 'clear',
        '\x1D': ['i', '_'],
        '\x1F': ['u', '-']
    }

    def __init__(self):
        self.log = logging.getLogger('ad1459.formatting')
    
    def format_text(self, text):
        for i in self.formatting:
            # text = text.replace(f'/{self.formatting[i][0]}', i)
            text = text.replace(f'/{self.formatting[i][1]}', i)
        return text
    
    def sanitize_message(self, message):
        """ Strips IRC formatting from a message to make it suitable for display
        outside of the client, e.g. in notifications.

        Arguments:
            message (str): The message to clean up.

        Returns:
            str: The cleaned message.
        """
        formatting_chars = [
            '\x02', # bold
            '\x1D', # italics
            '\x1F', # underline
            '\x1E', # strikethrough
            '\x11'
        ]

        for fchar in formatting_chars:
            message = message.replace(fchar, '')
        
        return message

    def parse_text(self, text):        
        # \u0002 bold
        # \u0003 colour
        # \u000F cancel all
        # \u001D italic
        # \u001F underline
        text = self.fix_markedup_tags(text)
        f_text = ''
        current_tags = []

        for char in text:
            if char in self.formatting:

                if not char in current_tags:
                    for tag in reversed(current_tags):
                        f_text = f'{f_text}</{self.formatting[tag][0]}>'

                    current_tags.append(char)
                    for tag in current_tags:
                        f_text = f'{f_text}<{self.formatting[tag][0]}>'

                else:
                    for tag in reversed(current_tags):
                        f_text = f'{f_text}</{self.formatting[tag][0]}>'

                    current_tags.pop()
                    for tag in current_tags:
                        f_text = f'{f_text}<{self.formatting[tag][0]}>'

            else: 
                f_text = f'{f_text}{char}'
        
        for tag in reversed(current_tags):
            f_text = f'{f_text}</{self.formatting[tag][0]}>'

        return f_text
    
    def hyperlinks(self, text):
        words = text.split()

        linked_words = []
        for word in words:
            scheme = urlparse(word).scheme
            if scheme == 'http' or scheme == 'https':
                paren = False
                if word.endswith(')'):
                    word = word.strip(')')
                    paren = True
                word = f'<a href="{word}">{word}</a>'
                if paren:
                    word = f'{word})'
            linked_words.append(word)
        return " ".join(linked_words)

    def fix_markedup_tags(self, text):
        mu_formatting = {
            '&#x2;': '\x02',
            #'\u0003': #'\u0003',
            #'\u000F': #'\u000F',
            '&#x1d;': '\x1D',
            '&#x1f;': '\x1F'
        }

        for i in mu_formatting:
            text = text.replace(i, mu_formatting[i])
        
        return text


#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  Main Application
"""

import asyncio
import logging
import pathlib
import os
import threading

import gi
gi.require_versions(
    {
        'Gtk': '3.0',
        'Gdk': '3.0'
    }
)
from gi.repository import Gtk, Gdk, Gio

from .formatting import Parser
from .network import Network
from .widgets.window import Ad1459Window
from . import handlers

USER_HOME_PATH = str(pathlib.Path.home())
CONFIG_DIR_PATH = os.path.join(
    USER_HOME_PATH, 
    os.path.join('.config', 'ad1459')
)
CONFIG_FILE_PATH = os.path.join(CONFIG_DIR_PATH, 'servers.ini')

class Ad1459Application:

    def __init__(self):
        self.log = logging.getLogger('ad1459')
        handler = logging.StreamHandler()
        formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
        handler.setFormatter(formatter)
        self.log.addHandler(handler)
        self.log.setLevel(logging.DEBUG)
        self.log.debug('Initializing application')

        self.user_home_path = USER_HOME_PATH
        self.config_dir_path = CONFIG_DIR_PATH
        self.config_file_path = CONFIG_FILE_PATH

        self.windows = []
        self.networks = []

        self.app = Gtk.Application.new(
            'in.donotspellitgav.in', Gio.ApplicationFlags.FLAGS_NONE
        )
        self.app.connect('activate', self.init_application)
        self.app.parser = Parser()
        self.parser = Parser()

                # Set up CSS
        css = (
            b'.message-row {'
            b'  border-radius: 6px;'
            b'  background-color: alpha(@theme_selected_bg_color, 0.1);'
            b'}'
            b'.mine {'
            b'  background-color: alpha(@theme_selected_bg_color, 0.2);'
            b'}'
            b'.highlight {'
            b'    background-color: alpha(@success_color, 0.3);'
            b'}'
            b'.server {'
            b'    background-color: transparent;'
            b'}'
            b'.notice {'
            b'    background-color: alpha(@warning_color, 0.3);'
            b'}'
            b'.connect-entry {'
            b'    background-color: @theme_base_color;'
            b'}'
            b'.topic-expander {'
            b'    font-size: 0.75em;'
            b'}'
        )

        screen = Gdk.Screen.get_default()
        css_provider = Gtk.CssProvider()
        css_provider.load_from_data(css)
        style_context = Gtk.StyleContext()
        style_context.add_provider_for_screen(
            screen, css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def init_application(self):
        """ Starts up all of the application loops and peices."""
        self.log.debug('Initializing IRC thread')
        self.irc = threading.Thread(target=asyncio.get_event_loop().run_forever)
        self.irc.daemon = True
        self.irc.start()

        self.log.debug('Creating initial window...')
        first_window = self.init_window()
        self.windows.append(first_window)
        first_window.show_all()

        self.log.debug('Starting up GTK main loop.')
        Gtk.main()

    def init_window(self):
        """ Create a a window for the application.
        
        Returns:
            A Gtk.Window set up for AD1459
        """
        self.log.debug('Adding window')
        window = Ad1459Window(self)
        window.props.application = self.app
        window.set_default_size(1000, 600)
        window.connect('delete-event', self.remove_window)
        self.connect_ui(window)
        
        window.show_all()
        return window

    def connect_ui(self, window):
        """ Connects the UI in window to the correct signal handlers.

        Arguments:
            window (:obj:`Ad1459Windw`): The window object to connect.
        """
        # Hook up the IRC Entry and related buttons
        window.irc_entry.connect(
            'activate', 
            handlers.on_send_button_clicked,
            window.irc_entry.get_text(),
            window.switcher.get_active_room(),
            window
        )
        window.send_button.connect(
            'clicked', 
            handlers.on_send_button_clicked,
            window.irc_entry.get_text(),
            window.switcher.get_active_room(),
            window
        )
        window.nick_button.connect(
            'clicked',
            handlers.on_nick_button_clicked,
            window
        )

        # Hook up header functions
        window.header.close_button.connect(
            'clicked',
            handlers.on_appmenu_close_clicked,
            window.switcher.get_active_room(),
            window
        )
        # window.header.keys_button.connect(
        #     'clicked',
        #     handlers.on_appmenu_keys_clicked,
        #     window
        # )
        window.header.about_button.connect(
            'clicked',
            handlers.on_appmenu_about_clicked,
            window
        )

        # ServerPopup functions
        window.header.server_popup.connect_button.connect(
            'clicked',
            handlers.on_server_popup_connect_clicked,
            window
        )

        # Channel Switcher 
        window.switcher.switcher.connect(
            'row-selected',
            handlers.on_room_selected,
            window
        )
        window.join_entry.connect(
            'activate',
            handlers.on_join_entry_activate,
            window
        )
        window.join_entry.connect(
            'icon-release',
            handlers.on_join_entry_icon_release,
            window
        )

    def remove_window(self, window, data=None):
        """ Deletes a window and moves all of its stuff to another window.

        Arguments: 
            window (Gtk.Window): The window to remove.
        """
        self.log.debug('Deleting window')
        window.destroy()

        if len(self.windows) <= 1:
            self.log.debug('Last window destroyed, quitting.')
            Gtk.main_quit()
            exit(0)

    # def add_network(
    #         self,
    #         name=None,
    #         auth=None,
    #         host=None,
    #         port=None,
    #         tls=True,
    #         nick=None,
    #         user=None,
    #         real=None,
    #         pasw=None
    # ):
    #     """ Adds a network object to this application.

    #     Returns:
    #         A :obj:`Network` for the new network.
    #     """
    #     new_network = Network(self)
        
    #     if name:
    #         new_network.name = name

    #     if auth:
    #         new_network.auth = auth

    #     if host:
    #         new_network.host = host

    #     if port:
    #         new_network.port = port

    #     if not tls:
    #         new_network.tls = tls

    #     if nick:
    #         new_network.nickname = nick

    #     if user:
    #         new_network.username = user

    #     if real:
    #         new_network.realname = real

    #     if pasw:
    #         new_network.password = pasw



#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  Signal handlers for UI components.
"""

import logging

from .network import Network
from .widgets.room_row import RoomKind
from .widgets.about import AboutDialog

log = logging.getLogger('ad1459.handers')

def on_send_button_clicked(widget, text, room, window, data=None):
    """ `clicked` signal handler for the Send button.

    Also handles the `activate` signal for the IRC Entry.
    
    Arguments:
        text (str): The text to send to IRC.
        room (:obj:`Room`): The room to send `text` to.
        window (:obj:`Ad1459Application`): The window we're in
    """
    log.debug('on_send_button_clicked handler triggered')
    room = window.message_stack.get_visible_child().room
    network = room.network
    message = window.irc_entry.get_text()

    # We want to avoid sending empty messages
    if message:
        network.send_message(room, message)

    window.irc_entry.set_text('')
    room.unsent_text = ''
    room.recents.append(message)
    window.recents.append(message)

    # Keep the recent sent messages to a reasonable size.
    while len(room.recents) > 50:
        room.recents.pop(0)
    while len(window.recents) > 50:
        room.recents.pop(0)
    

def on_nick_button_clicked(button, window, data=None):
    """`clicked` signal handler for the nickname button.
    
    Arguments:
        text (str): The text to set the nickname to.
        network (:obj:`Network`): The network whose nick to set.
        window (:obj:`Ad1459Application`): The window we're in
    """
    network = window.active_room.network
    new_nick = window.irc_entry.get_text()
    network.change_nick(new_nick)
    window.irc_entry.set_text('')

def on_server_popup_connect_clicked(button, window, data=None):
    """`clicked` signal handler for the Connect button in the server popup
    
    Arguments:
        window (:obj:`Ad1459Application`): The window we're in
    """
    popup = window.header.server_popup
    app = window.app
    network_line = popup.server_line
    window.header.spinner.start()
    
    if network_line:
        pass

    else:
        network = Network(app, window)
        network.name = popup.name
        network.auth = popup.auth
        network.host = popup.server
        network.port = popup.port
        network.tls = popup.tls
        network.nickname = popup.nick
        network.username = popup.username
        network.realname = popup.realname
        network.password = popup.password
    
    app.networks.append(network)
    
    if popup.save:
        popup.save_details()

    network.connect()
    popup.popdown()
    popup.layout_grid.set_sensitive(False)

def on_appmenu_close_clicked(button, room, window, data=None):
    """`clicked` signal handler for the Leave button in the appmenu.
    
    Arguments:
        room (:obj:`Room`): The Room to leave.
        window (:obj:`Ad1459Application`): The window we're in
    """
    room = window.message_stack.get_visible_child().room
    network = room.network
    if room.kind == RoomKind.CHANNEL:
        room.part()
    elif room.kind == RoomKind.SERVER:
        for room in network.rooms:
            room.leave()
        room.network.disconnect()
        window.app.networks.remove(network)
        window.switcher.invalidate_sort()
    else:
        room.network.rooms.remove(room)
        room.leave()
        window.switcher.invalidate_sort()
    
    window.switcher.switcher.select_row(network.server_room.row)

def on_appmenu_about_clicked(button, window, data=None):
    """`clicked` signal handler for the About button.
    
    Arguments:
        window (:obj:`Ad1459Application`): The window we're in
    """
    about_dialog = AboutDialog()
    about_dialog.run()
    about_dialog.destroy()

def on_room_selected(listbox, row, window, data=None):
    """`row-selected` signal handler for room switcher.

    Arguments:
        window (:obj:`Gtk.Window`): The window we're part of.
    """
    try:
        row.room.update_users()
        row.room.topic_pane.update_topic()
        row.room.notification.close()
        window.show_all()
        row.icon = 'none'
        window.message_stack.set_visible_child_name(row.room.id)
        window.topic_stack.set_visible_child_name(row.room.id)
        window.irc_entry.grab_focus_without_selecting()
        window.nick_button.set_label(row.room.network.nickname)
        window.irc_entry.set_text(row.room.unsent_text)
        length = len(row.room.unsent_text)
        window.irc_entry.set_position(length)
    
        if row.room.kind == RoomKind.DIALOG:
            window.header.close_button.set_label('Close conversation')
        
        elif row.room.kind == RoomKind.SERVER:
            window.header.close_button.set_label('Disconnect from server')
        
        elif row.room.kind == RoomKind.CHANNEL:
            window.header.close_button.set_label('Leave channel')
    
    except AttributeError:
        return
    
    window.switcher.invalidate_sort()

def on_join_entry_activate(entry, window, data=None):
    """`activate` signal handler for join entry.

    also handles the `icon-release` signal.
    """
    room_name = entry.get_text()
    network = window.message_stack.get_visible_child().room.network
    
    if room_name.startswith('#'):
        network.join_channel(room_name)
    
    else:
        room = network.get_room_for_name(room_name)
    
    entry.set_text('')

def on_join_entry_icon_release(entry, icon_pos, event, window, data=None):
    """`activate` signal handler for join entry.

    also handles the `icon-release` signal.
    """
    room_name = entry.get_text()
    network = window.message_stack.get_visible_child().room.network
    
    if room_name.startswith('#'):
        network.join_channel(room_name)
    
    else:
        room = network.get_room_for_name(room_name)
    
    entry.set_text('')



#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  This file is an application window.
"""

import logging

import gi
gi.require_versions(
    {
        'Gtk': '3.0',
        'Gdk': '3.0'
    }
)
from gi.repository import Gtk, Gdk
from .room_row import room_row_sort

class RoomSwitcher(Gtk.Grid):
    """ The room switcher sidebar content in the main window. 

    This contains room rows for all of the currently open rooms, and allows the
    user to click on them to switch to a different room.
    """

    def __init__(self, app):
        super().__init__()
        self.app = app

        window = Gtk.ScrolledWindow()
        window.set_hexpand(True)
        window.set_vexpand(True)
        self.attach(window, 0, 0, 1, 1)

        self.switcher = Gtk.ListBox()
        self.switcher.set_hexpand(True)
        self.switcher.set_vexpand(True)
        self.switcher.set_sort_func(room_row_sort)
        window.add(self.switcher)
    
    def add_row(self, row):
        """ Adds a row to the ListBox.

        Arguments:
            row (:obj:`RoomRow`): A Room Row to add to the listbox.
        """
        self.switcher.insert(row, 0)
    
    def remove_row(self, row):
        """ Removes a row from the ListBox.

        Arguments:
            row (:obj:`RoomRow`): The row to remove
        """
        try:
            row.destroy()

        except AttributeError:
            pass

        self.switcher.invalidate_sort()
    
    def invalidate_sort(self):
        """ proxy for self.switcher.invalidate_sort()"""
        self.switcher.invalidate_sort()
    
    def get_active_room(self):
        """Gets the currently selected Room.

        Returns:
            a :obj:`Room` for the currently selected row
        """
        room = self
        room.network = ""
        return room




#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  A popover for connecting to servers.
"""

import configparser
import keyring as Keyring
import os
import pathlib

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk

class ServerPopover(Gtk.Popover):
    """ A popover for connecting to a server.

    This hooks into our server details storage to look up saved configurations.
    It also allows for connecting to new networks or modifying existing ones. It
    will eventually also allow deleting networks from the list of saved networks.

    This class uses the text of its entries as a data store for the below 
    attributes.

    Attributes:
        name (str): The name of the server.
        nick (str): The nickname to use on this network.
        username (str): The username/ident to give to the user.
        realname (str): The realname to display in WHOIS.
        server (str): The server for this connection.
        port (int): The port on `server` to connect to.
        auth (str): The authentication to use. One of `pass`, `sasl`, or `none`.
        password (str): The password to supply for authentication, if required.
        tls (bool): Whether this connection uses TLS.
        save (bool): Whether to save this connection.
        server_line (str): A condensed line to enter the connection in one-shot.
    """

    def __init__(self, config_file_path):
        super().__init__()

        self.config = configparser.ConfigParser()
        self.keyring = Keyring.get_keyring()

        self.config_file_path = config_file_path
        pathlib.Path(os.path.dirname(self.config_file_path)).mkdir(
            parents=True, exist_ok=True
        )

        self.config.read(self.config_file_path)
        
        self.layout_grid = Gtk.Grid()
        self.layout_grid.set_column_spacing(6)
        self.layout_grid.set_row_spacing(12)
        self.layout_grid.props.margin = 6
        self.add(self.layout_grid)
        
        # Saved Networks
        self.saved_combo = Gtk.ComboBoxText()
        self.init_saved_combo()
        self.saved_combo.set_active(-1)
        Gtk.StyleContext.add_class(
            self.saved_combo.get_style_context(), 'connect-entry'
        )
        self.saved_combo.connect('changed', self.on_saved_combo_changed)
        self.layout_grid.attach(self.saved_combo, 0, 0, 2, 1)

        self.connect_grid = Gtk.Grid()
        self.connect_grid.set_orientation(Gtk.Orientation.VERTICAL)
        Gtk.StyleContext.add_class(self.connect_grid.get_style_context(), 'linked')
        Gtk.StyleContext.add_class(self.connect_grid.get_style_context(), 'connect-entry')
        self.layout_grid.attach(self.connect_grid, 0, 1, 2, 1)

        # Network Name
        self.name_entry = Gtk.Entry()
        self.name_entry.simple_entry = True
        self.name_entry.set_placeholder_text('Network Name')
        self.connect_grid.attach(self.name_entry, 0, 0, 2, 1)

        # nickname
        self.nick_entry = Gtk.Entry()
        self.nick_entry.simple_entry = True
        self.nick_entry.set_placeholder_text('Nickname')
        self.connect_grid.attach(self.nick_entry, 0, 1, 2, 1)

        # username
        self.username_entry = Gtk.Entry()
        self.username_entry.simple_entry = True
        self.username_entry.set_placeholder_text('Username/Ident')
        self.connect_grid.attach(self.username_entry, 0, 2, 2, 1)

        # realname
        self.realname_entry = Gtk.Entry()
        self.realname_entry.simple_entry = True
        self.realname_entry.set_placeholder_text('Real Name')
        self.connect_grid.attach(self.realname_entry, 0, 3, 2, 1)

        # server
        self.server_entry = Gtk.Entry()
        self.server_entry.simple_entry = True
        self.server_entry.set_placeholder_text('Server')
        self.connect_grid.attach(self.server_entry, 0, 4, 1, 1)

        # port
        port_adj = Gtk.Adjustment.new(
            value=6697,
            lower=1,
            upper=65535,
            step_increment=1,
            page_increment=1,
            page_size=0
        )

        self.port_entry = Gtk.SpinButton()
        self.port_entry.simple_entry = True
        self.port_entry.configure(port_adj, 1, 0)
        self.port_entry.set_numeric(True)
        self.port_entry.set_placeholder_text('Port')
        self.connect_grid.attach(self.port_entry, 1, 4, 1, 1)

        # auth
        self.auth_combo = Gtk.ComboBoxText()
        self.auth_combo.simple_entry = True
        self.auth_combo.append_text('SASL Authentication')
        self.auth_combo.append_text('Server Password')
        self.auth_combo.append_text('No Authentication')
        self.auth_combo.set_active(0)
        Gtk.StyleContext.add_class(
            self.auth_combo.get_style_context(), 'connect-entry'
        )
        self.connect_grid.attach(self.auth_combo, 0, 5, 2, 1)

        # password
        self.password_entry = Gtk.Entry()
        self.password_entry.simple_entry = True
        self.password_entry.set_placeholder_text('Password')
        self.password_entry.set_visibility(False)
        self.connect_grid.attach(self.password_entry, 0, 6, 2, 1)

        # TLS Check
        self.tls_check = Gtk.CheckButton()
        self.tls_check.simple_entry = True
        self.tls_check.set_label('TLS')
        self.tls_check.set_active(True)
        self.layout_grid.attach(self.tls_check, 0, 2, 1, 1)

        # Save check
        self.save_check = Gtk.CheckButton()
        self.save_check.set_label('Save')
        self.save_check.set_active(True)
        self.layout_grid.attach(self.save_check, 1, 2, 1, 1)

        # Server entry
        self.server_line_entry = Gtk.Entry()
        self.server_line_entry.simple_entry = False
        self.server_line_entry.set_placeholder_text('Server line')
        self.layout_grid.attach(self.server_line_entry, 0, 3, 2, 1)

        # Connect Button
        self.connect_button = Gtk.Button()
        self.connect_button.set_label('Connect')
        Gtk.StyleContext.add_class(
            self.connect_button.get_style_context(), 'suggested-action'
        )
        self.layout_grid.attach(self.connect_button, 0, 4, 2, 1)

        self.widgets = [
            self.name_entry,
            self.nick_entry,
            self.username_entry,
            self.realname_entry,
            self.server_entry,
            self.password_entry
        ]

    # Methods
    def init_saved_combo(self):
        """Clears any existing items from the saved network combo and sets it 
        up with new ones.
        """
        self.saved_combo.remove_all()
        self.saved_combo.append_text('New...')
        for network in self.config.sections():
            self.saved_combo.append_text(network)
    
    def save_details(self):
        self.config[self.name] = {}
        self.config[self.name]['name'] = self.name
        self.config[self.name]['auth'] = self.auth
        self.config[self.name]['host'] = self.server
        self.config[self.name]['port'] = str(int(self.port))
        self.config[self.name]['tls'] = str(self.tls)
        self.config[self.name]['nickname'] = self.nick
        self.config[self.name]['username'] = self.username
        self.config[self.name]['realname'] = self.realname

        with open(self.config_file_path, mode='w') as configfile:
            self.config.write(configfile)
        
        keyring_username = f'{self.nick}!{self.username}@{self.name}'
        self.keyring.set_password(
            f'{self.name}-{self.server}',
            keyring_username,
            self.password
        )

    def reset_all_text(self):
        self.saved_combo.set_active(0)
        self.name_entry.set_text('')
        self.nick_entry.set_text('')
        self.username_entry.set_text('')
        self.realname_entry.set_text('')
        self.server_entry.set_text('')
        self.port_entry.set_value(6697)
        self.auth_combo.set_active(0)
        self.password_entry.set_text('')
        self.tls_check.set_active(True)
        self.save_check.set_active(True)
        self.server_line_entry.set_text('')

    def get_all_widgets_text(self):
        """ Returns the text of all widgets, formatted as a serverline."""

        line = ''
        if self.auth_combo.get_active_text() == 'SASL Authentication':
            line = 'sasl'
            
        elif self.auth_combo.get_active_text() == 'Server Password':
            line = 'pass'

        else:
            line = 'none'

        line = f'{line} {self.name_entry.get_text()}'
        line = f'{line} {self.server_entry.get_text()}'
        line = f'{line} {int(self.port_entry.get_value())}'
        line = f'{line} {self.nick_entry.get_text()}'
        line = f'{line} {self.username_entry.get_text()}'
        line = f'{line} {self.realname_entry.get_text()}'
        line = f'{line} {self.password_entry.get_text()}'
        
        return line.strip()

    # Internal Handlers
    def on_saved_combo_changed(self, combo, data=None):
        """ changed signal handler for saved_combo."""
        network = combo.get_active_text()

        if network == 'New...':
            self.reset_all_text()
        
        else:
            self.config.read(self.config_file_path)
            self.name_entry.set_text(
                self.config[network]['name']
            )
            self.nick_entry.set_text(
                self.config[network]['nickname']
            )
            self.username_entry.set_text(
                self.config[network]['username']
            )
            self.realname_entry.set_text(
                self.config[network]['realname']
            )
            self.server_entry.set_text(
                self.config[network]['host']
            )
            self.port_entry.set_value(
                int(self.config[network]['port'])
            )
            if self.config[network]['auth'] == 'sasl':
                self.auth_combo.set_active(0)
            elif self.config[network]['auth'] == 'pass':
                self.auth_combo.set_active(1)
            else:
                self.auth_combo.set_active(2)
            if self.config[network]['tls'] == 'True':
                self.tls_check.set_active(True)
            else:
                self.tls_check.set_active(False)
            keyring_service = (
                f'{self.config[network]["name"]}-{self.config[network]["host"]}'
            )
            keyring_username = (
                f'{self.config[network]["nickname"]}'
                f'!{self.config[network]["username"]}'
                f'@{self.config[network]["name"]}'
            )
            self.password_entry.set_text(
                self.keyring.get_password(keyring_service, keyring_username)
            )
            self.save_check.set_active(True)

    # Data
    @property 
    def name(self):
        return self.name_entry.get_text()

    @property 
    def nick(self):
        return self.nick_entry.get_text()

    @property 
    def username(self):
        if self.username_entry.get_text() != '':
            return self.username_entry.get_text()
        else:
            return self.nick

    @property 
    def realname(self):
        if self.realname_entry.get_text() != '':
            return self.realname_entry.get_text()
        else:
            return self.nick

    @property 
    def server(self):
        return self.server_entry.get_text()

    @property 
    def port(self):
        return self.port_entry.get_value()

    @property 
    def auth(self):
        auth = self.auth_combo.get_active_text()
        if auth == 'SASL Authentication':
            return 'sasl'

        elif auth == 'Server Password':
            return 'pass'

        else:
            return 'none'

    @property 
    def password(self):
        return self.password_entry.get_text()

    @property 
    def tls(self):
        return self.tls_check.get_active()

    @property 
    def save(self):
        return self.save_check.get_active()

    @property 
    def server_line(self):
        return self.server_line_entry.get_text()



#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  A buffer full of messages.
"""

import logging

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

class MessageBuffer(Gtk.ScrolledWindow):
    """ A buffer to display messages.

    This should be added as a child of a Gtk.Stack in the main window.
    
    Attributes:
        room (`Room`): The room this buffer belongs to.
    """

    def __init__(self, room):
        super().__init__()
        self.room = room

        self.list_box = Gtk.ListBox()
        self.list_box.set_vexpand(True)
        self.list_box.set_hexpand(True)
        self.list_box.connect('size-allocate', self.scroll_to_bottom)

        self.adj = self.get_vadjustment()
        self.adj.connect('value-changed', self.on_window_scroll)
        self.add(self.list_box)

    def add_message_to_buffer(self, message):
        """Adds a message into the buffer."""
        self.list_box.add(message)

        # Scroll to the bottom if this is our own message.
        if message.sender == self.room.network.nickname:
            self.scroll_to_bottom(self.list_box)
    
    def on_window_scroll(self, widget, data=None):
        """value-changed signal handler for the window adjustment.
        
        This is used to detect when the user scrolls up, so that we can 
        disconnect the `scroll_to_bottom` handler when we get a new message.
        """
        adj = self.get_vadjustment()
        max_value = adj.get_upper() - adj.get_page_size()

        if adj.get_value() < max_value:
            try:
                self.list_box.disconnect_by_func(self.scroll_to_bottom)
            
            except TypeError:
                pass
        
        else:
            try:
                self.list_box.disconnect_by_func(self.scroll_to_bottom)
            
            except TypeError:
                pass

            self.list_box.connect('size-allocate', self.scroll_to_bottom)

    def scroll_to_bottom(self, widget, data=None):
        """Scroll to the bottom of the buffer."""
        adj = self.get_vadjustment()
        max_value = adj.get_upper() - adj.get_page_size()
        adj.set_value(max_value)


#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  ListBoxRows for networks/rooms.
"""

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

class UserRow(Gtk.ListBoxRow):
    """ A row for a user in the user list.

    The attributes for this class are stored in the label widget for this row.

    Attributes:
        nick (str): This user's nickname.
    """

    def __init__(self, room):
        super().__init__()
        self.room = room
        grid = Gtk.Grid()
        grid.set_column_spacing(6)
        grid.set_margin_top(3)
        grid.set_margin_bottom(3)
        grid.set_margin_start(3)
        grid.set_margin_end(3)
        self.add(grid)

        self.status_image = Gtk.Image.new_from_icon_name(
            'radio-symbolic',
            Gtk.IconSize.SMALL_TOOLBAR
        )
        self.status_image.props.opacity = 0
        grid.attach(self.status_image, 0, 0, 1, 1)

        self.user_label = Gtk.Label()
        grid.attach(self.user_label, 1, 0, 1, 1)

        self._modes = []
    
    # Data
    @property
    def nick(self):
        """str: The user's nickname."""
        return self.user_label.get_text()
    
    @nick.setter
    def nick(self, nick):
        self.user_label.set_text(nick)

    @property
    def modes(self):
        """:list: of str: The current user modes for this user"""
        return self._modes
    
    @modes.setter
    def modes(self, modes):
        self._modes = modes
        self.status_image.set_opacity(0)

        if 'q' in self._modes:
            self.status_image.set_from_icon_name(
                'view-conceal-symbolic',
                Gtk.IconSize.SMALL_TOOLBAR
            )
            self.status_image.set_opacity(0.5)
        
        if 'v' in self._modes:
            self.status_image.set_from_icon_name(
                'media-record-symbolic',
                Gtk.IconSize.SMALL_TOOLBAR
            )
            self.status_image.set_opacity(1)

        if 'o' in self._modes:
            self.status_image.set_from_icon_name(
                'starred-symbolic',
                Gtk.IconSize.SMALL_TOOLBAR
            )
            self.status_image.set_opacity(1)
    
    @property
    def op(self):
        """bool: True if user is chanop."""
        if 'o' in self.modes:
            return True
        
    @property
    def voice(self):
        if 'v' in self.modes and not 'o' in self.modes:
            return True
    
    @property
    def mute(self):
        if 'q' in self.modes and not 'o' in self.modes and not 'v' in self.modes:
            return True
    
    @property
    def nomodes(self):
        if not 'q' in self.modes and not 'o' in self.modes and not 'v' in self.modes:
            return True



#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  Handling for networks and their rooms.
"""

import asyncio
import logging
import time

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib

from .widgets.room_row import RoomRow
from .widgets.message_row import MessageRow
from .client import Client
from .commands import Commands
from .formatting import Parser
from .room import Room

class Network:
    """An IRC network to connect to and chat on.

    Attributes:
        rooms (list of :obj:`Room`): A list of the joined rooms on this network.
        name (str): The user-defined name for this network.
        auth (str): The user-authentication for this network. 'sasl', 'pass', 
            or 'none'.
        host (str): The hostname for the server to connect to.
        port (int): The port number to use.
        tls (bool): Wheter the connection uses TLS encryption.
        nickname (str): The user's nickname on this network.
        realname (str): The user's Real Name on this network.
        password (str): The authentication for the password on this network.
        app (:obj:`Ad1459Application`): The app we're running on.
    """
    cmd = Commands()
    commands = {
        '/me': cmd.me,
        '/join': cmd.join,
        '/quit': cmd.quit,
        '/part': cmd.part,
        '/kick': cmd.kick,
        '/ban': cmd.ban,
        '/kickban': cmd.kickban,
        # '/cycle': cmd.cycle,
        '/nick': cmd.nick,
        '/query': cmd.query,
        '/msg': cmd.msg,
        '/whisper': cmd.query,
        '/notice': cmd.notice,
        '/topic': cmd.topic
    }

    def __init__(self, app, window):
        self.log = logging.getLogger('ad1459.network')
        self.log.debug('Creating network')
        self.app = app
        self.parser = Parser()
        self.window = window
        self.rooms = []
        self._config = {
            'name': 'New Network',
            'auth': 'sasl',
            'host': 'chat.freenode.net',
            'port': 6697,
            'tls': True,
            'nickname': 'ad1459-user',
            'username': 'ad1459-user',
            'realname': 'AD1459 User',
            'password': 'hunter2'
        }
        self.client = None

    # Synchronous Methods for this object.
    def connect(self):
        """ Connect to the network."""
        if self.client:
            if self.client.connected:
                self.log.info('Disconnecting from %s', self.host)
                self.disconnect()

        elif self.auth == 'sasl':
            self.client = Client(
                self.nickname, 
                self, 
                sasl_password=self.password, 
                sasl_username=self.username
            )
        else:
            self.client = Client(self.nickname, self)
        
        self.log.debug('Spinning up async connection to %s', self.host)
        self.client.username = self.username
        self.server_room = Room(self.app, self, self.window, self.name)
        self.server_room.kind = "server"
        self.add_room(self.server_room)

        # Actually do the connection
        if self.auth == 'pass':
            self.log.debug('Using password authentication')
            asyncio.run_coroutine_threadsafe(
                self.client.connect(
                    hostname=self.host,
                    port=self.port,
                    tls=self.tls,
                    password=self.password
                ),
                loop=asyncio.get_event_loop()
            )

        else:
            self.log.debug('Using SASL authentication (or none)')
            asyncio.run_coroutine_threadsafe(
                self.client.connect(
                    hostname=self.host,
                    port=self.port,
                    tls=self.tls
                ),
                loop=asyncio.get_event_loop()
            )
    
    def disconnect(self, message='AD1459 Quit'):
        """Disconnect from this network."""
        self.log.debug('Disconnecting from %s', self.name)
        asyncio.run_coroutine_threadsafe(
            self.client.quit(message='AD1459 Quit'),
            loop=asyncio.get_event_loop()
        )
    
    def change_nick(self, new_nick):
        """ Changes the user's nick.
        
        Arguments:
            new_nick (str): The new nickname to use.
        """
        asyncio.run_coroutine_threadsafe(
            self.client.set_nickname(new_nick),
            loop=asyncio.get_event_loop()
        )

    def join_channel(self, channel):
        """ Joins a room on the network.
        
        Arguments:
            channel (str): The name of the channel to join.
        """
        asyncio.run_coroutine_threadsafe(
            self.client.join(channel), loop=asyncio.get_event_loop()
        )

    def add_room(self, room):
        """Adds a room to this network and adds it to the UI.
        
        Arguments:
            room (:obj:`Room`): The room object to add
        """
        self.log.debug('Adding room %s to the window', room.name)
        self.window.switcher.add_row(room.row)
        self.window.message_stack.add_named(room.buffer, room.id)
        self.window.topic_stack.add_named(room.topic_pane, room.id)
        self.window.show_all()
        self.window.switcher.invalidate_sort()
        self.rooms.append(room)
    
    def get_room_for_name(self, name):
        """ Gets a room object given the name.
        
        Arguments:
            name (str): The name of the room to join
        """
        for room in self.rooms:
            if room.name == name:
                return room

        if not name.startswith('#'):
            new_room = Room(self.app, self, self.window, name)
            new_room.kind = 'dialog'
            new_room.name = name
            new_room.topic_pane.update_topic()
            self.add_room(new_room)
            self.window.switcher.invalidate_sort()
            return new_room
    
    def send_message(self, room, message):
        """ Sends a message/PRIVMSG to a channel/user.
        
        Arguments:
            room (:obj:`Room`): The room to send the message to. This is a 
                PM room or a channel.
            message (str): The message to send
        """
        message_text = self.parser.format_text(message)
        for command in self.commands:
            if not message.startswith(command):
                continue
            else:
                message_text = message_text.replace(command, '')
                self.log.debug('Got command: %s (%s)', command, message_text)
                self.commands[command](room, self, message_text)
                return 0

        self.log.debug('Sending message to %s: %s', room.name, message_text)
        asyncio.run_coroutine_threadsafe(
            self.client.message(room.name, message_text),
            loop=asyncio.get_event_loop()
        )
    
    # Asynchronous Callbacks
    async def on_connected(self):
        """ Called upon connection to IRC."""
        self.log.info('Connected to %s', self.name)
        GLib.idle_add(self.do_connected)
    
    def do_connected(self):
        popup = self.window.header.server_popup
        popup.reset_all_text()
        popup.layout_grid.set_sensitive(True)
        self.window.header.spinner.stop()
        self.window.switcher.invalidate_sort()
    
    async def on_nick_change(self, old, new):
        self.log.debug('Nick %s changed to %s', old, new)
        GLib.idle_add(self.do_nick_change, old, new)
    
    def do_nick_change(self, old, new):
        if old == self.nickname or old == '<unregistered>':
            self.nickname = new
            self.window.nick_button.set_label(self.nickname)
        
        else:
            for room in self.rooms:
                if new in room.users:
                    room.add_message(f'{old} is now {new}.', kind='server')
                    room.window.show_all()
                    room.update_users()
    
    async def on_join(self, channel, user):
        self.log.debug('%s has joined %s', user, channel)
        GLib.idle_add(self.do_join, channel, user)
    
    def do_join(self, channel, user):
        if user == self.nickname:
            new_channel = Room(self.app, self, self.window, channel)
            new_channel.kind = 'channel'
            new_channel.name = channel
            new_channel.topic_pane.update_topic()
            self.add_room(new_channel)
            new_channel.update_users()
            self.window.switcher.invalidate_sort()
            self.window.switcher.switcher.select_row(new_channel.row)
        
        else:
            room = self.get_room_for_name(channel)
            room.add_message(f'{user} joined.', kind='server')
            self.window.show_all()
            room.update_users()
    
    async def on_part(self, channel, user, message=None):
        self.log.debug('%s has left %s, (%s)', user, channel, message)
        GLib.idle_add(self.do_part, channel, user, message)
    
    def do_part(self, channel, user, message=None):
        room = self.get_room_for_name(channel)
        if user == self.nickname:
            room.leave()
            room.network.server_room.add_message(
                f'You have left {room.name}', kind='server'
            )
            self.rooms.remove(room)
            self.window.switcher.invalidate_sort()
        
        else:
            room.update_users()
            room.add_message(f'{user} left ({message}).', kind='server')
            self.window.show_all()
    
    async def on_kick(self, channel, target, by, reason=None):
        GLib.idle_add(self.do_kick, channel, target, by, reason)
    
    def do_kick(self, channel, target, by, reason=None):
        room = self.get_room_for_name(channel)
        if target == self.nickname:
            room.leave()
            room.network.server_room.add_message(
                f'You were kicked from {room.name} by {by}, reason: "{reason}"', 
                kind='server'
            )
            room.network.server_room.row.set_icon('emblem-important-symbolic')
            self.rooms.remove(room)
            self.window.switcher.invalidate_sort()
        
        else:
            room.update_users()
            room.add_message(
                f'{target} kicked by {by},({reason}).', kind='server'
            )
            self.window.show_all()

    async def on_quit(self, user, message=None):
        self.log.debug('%s has quit! (%s)', user, message)
        GLib.idle_add(self.do_quit, user, message)
    
    def do_quit(self, user, message=None):
        qmessage = f'{user} quit ({message}).'
        for room in self.rooms:
            if user in room.old_users:
                room.add_message(qmessage, kind='server')
                room.update_users()
    
    async def on_kill(self, target, by, reason=None):
        GLib.idle_add(self.do_kill, target, by, reason)
    
    def do_kill(self, target, by, reason=None):
        qmessage = f'{target} killed by {by}, ({reason}).'
        for room in self.rooms:
            if target in room.old_users:
                room.add_message(qmessage, kind='server')
                room.update_users()
    
    async def on_message(self, target, source, message):
        GLib.idle_add(self.do_message, target, source, message)
    
    def do_message(self, target, source, message):
        if target.startswith('#'):
            room = self.get_room_for_name(target)
            self.log.debug('Adding message to %s', room.id)
            room.add_message(message, source)
            room.update_tab_complete(source)
            self.window.show_all()

    async def on_notice(self, target, source, message):
        self.log.debug('%s noticed to %s: %s', source, target, message)
        GLib.idle_add(self.do_notice, target, source, message)
    
    def do_notice(self, target, source, message):
        if target.startswith('#'):
            room = self.get_room_for_name(target)
            self.log.debug('Adding notice to %s', room.id)
            room.add_message(message, source, kind='notice')
            room.update_tab_complete(source)
            self.window.show_all()
    
    async def on_private_message(self, target, source, message):
        self.log.debug('PM to %s from %s: %s', target, source, message)
        GLib.idle_add(self.do_private_message, target, source, message)

    def do_private_message(self, target, source, message):
        if target == self.nickname:
            room = self.get_room_for_name(source)
            self.log.debug('Adding message from %s', room.id)
        else:
            room = self.get_room_for_name(target)
            self.log.debug('Adding message to %s', room.id)
        
        room.add_message(message, source)
        self.window.show_all()
    
    async def on_private_notice(self, target, source, message):
        self.log.debug('Private Notice to %s from %s: %s', target, source, message)
        GLib.idle_add(self.do_private_notice, target, source, message)
        
    def do_private_notice(self, target, source, message):
        if target == self.nickname:
            room = self.get_room_for_name(source)
            self.log.debug('Adding message from %s', room.id)
        else:
            room = self.get_room_for_name(target)
            self.log.debug('Adding message to %s', room.id)

        self.log.debug('Adding notice from %s', room.id)
        room.add_message(message, source, kind='notice')
        self.window.show_all()
    
    async def on_topic_change(self, channel, message, by):
        GLib.idle_add(self.do_topic_change, channel, message, by)
    
    def do_topic_change(self, channel, message, by):
        self.log.debug('Changing topic in %s', channel)
        room = self.get_room_for_name(channel)
        room.topic_pane.update_topic()
        room.add_message(
            f'{by} set the topic to "{message}"', kind='server'
        )
    
    async def on_mode_change(self, channel, modes, by):
        GLib.idle_add(self.do_mode_change, channel, modes, by)

    def do_mode_change(self, channel, modes, by):
        room = self.get_room_for_name(channel)
        mode_codes = modes[0]
        mode_index = 1
        change = None
        last_code = ''
        mode_msg = f'{by} '
        try:
            for code in mode_codes:
                if code == last_code:
                    mode_msg += f'{modes[mode_index]}, '
                    mode_index += 1
                
                elif code == '+':
                    change = 'set'
                
                elif code == '-':
                    change = 'removed'
                
                elif code == 'b':
                    if change == 'set':
                        mode_msg += f'banned {modes[mode_index]}, '
                        mode_index += 1
                    else:
                        mode_msg += f'unbanned {modes[mode_index]}, '
                        mode_index += 1

                
                elif code == 'c':
                    if change == 'set':
                        mode_msg += 'disabled colours, '
                    else:
                        mode_msg += 'enabled colours, '
                
                elif code == 'e':
                    mode_msg += f'{change} a ban exemption on {modes[mode_index]}, '
                    mode_index += 1
                
                elif code == 'C':
                    if change == 'set':
                        mode_msg += 'disabled CTCP, '
                    else:
                        mode_msg += 'enabled CTCP, '
                
                elif code == 'i':
                    if change == 'set':
                        mode_msg += f'made {channel} invite-only, '
                    else:
                        mode_msg += f'made {channel} open, '

                elif code == 'o':
                    if change == 'set':
                        mode_msg += f'opped {modes[mode_index]}, '
                        mode_index += 1
                    else:
                        mode_msg += f'deopped {modes[mode_index]}, '
                        mode_index += 1
                
                elif code == 'v':
                    if change == 'set':
                        mode_msg += f'voiced {modes[mode_index]}, '
                        mode_index += 1
                    else:
                        mode_msg += f'devoiced {modes[mode_index]}, '
                        mode_index += 1
                
                elif code == 'q':
                    if change == 'set':
                        mode_msg += f'muted {modes[mode_index]}, '
                        mode_index += 1
                    else:
                        mode_msg += f'unmuted {modes[mode_index]}, '
                        mode_index += 1
                
                elif code == 'm':
                    mode_msg += f'{change} moderated, '

                last_code = code
        
        except IndexError:
            print('No index %s', mode_index)
        
        mode_msg = mode_msg.strip()
        mode_msg = mode_msg.strip(',')
        room.add_message(f'{mode_msg}.', kind='server')
        room.update_users()

    async def on_user_invite(self, target, channel, by):
        GLib.idle_add(self.do_user_invite, target, channel, by)
    
    def do_user_invite(self, target, channel, by):
        room = self.get_room_for_name(channel)
        room.add_message(f'{by} invited {target}', kind='server')

    async def on_invite(self, channel, by):
        GLib.idle_add(self.do_invite, channel, by)
    
    def do_invite(self, channel, by):
        self.server_room.add_message(
            f'You were invited to {channel} by {by}', kind='server'
        )
        self.server_room.row.set_icon('emblem-important-symbolic')


    async def on_ctcp_action(self, target, source, action):
        self.log.debug('Action in %s from %s: %s %s', target, source, source, action )
        GLib.idle_add(self.do_ctcp_action, target, source, action)
        
    def do_ctcp_action(self, target, source, action):
        self.log.debug('Adding action from %s to %s', source, target)
        if target.startswith('#'):
            room = self.get_room_for_name(target)
        elif target == self.nickname:
            room = self.get_room_for_name(source)
        
        message = f'\x1D{source} {action}\x1D'
        room.add_message(message, kind='action')
        self.window.show_all()
        room.update_tab_complete(source)

    # Data for this object.
    @property
    def name(self):
        """str: The name of this network (and its room)."""
        return self._config['name']
    
    @name.setter
    def name(self, name):
        """This is actually tracked by the room."""
        self.log.debug('Setting name to %s', name)
        self._config['name'] = name
    
    @property
    def auth(self):
        """str: One of 'sasl', 'pass', or 'none'."""
        return self._config['auth']
    
    @auth.setter
    def auth(self, auth):
        """Only set if it's a valid value."""
        if auth == 'sasl' or auth == 'pass' or auth == 'none':
            self._config['auth'] = auth

    @property
    def host(self):
        """str: The hostname of the server to connect to."""
        return self._config['host']
    
    @host.setter
    def host(self, host):
        self._config['host'] = host
    
    @property
    def port(self):
        return self._config['port']
    
    @port.setter
    def port(self, port):
        """ Only set a port that is within the valid range."""
        if port > 0 and port <= 65535:
            self._config['port'] = int(port)

    @property
    def tls(self):
        """bool: Whether or not to use TLS"""
        return self._config['tls']
    
    @tls.setter
    def tls(self, tls):
        self._config['tls'] = tls

    @property
    def nickname(self):
        """str: The user's nickname"""
        return self._config['nickname']
    
    @nickname.setter
    def nickname(self, nickname):
        self.log.debug('Setting nickname to %s', nickname)
        self._config['nickname'] = nickname

    @property
    def username(self):
        """str: The username to use for the connection"""
        return self._config['username']
    
    @username.setter
    def username(self, username):
        self.log.debug('Setting username to %s', username)
        self._config['username'] = username

    @property
    def realname(self):
        """str: The user's real name"""
        return self._config['realname']
    
    @realname.setter
    def realname(self, realname):
        self._config['realname'] = realname

    @property
    def password(self):
        """str: The user's password."""
        return self._config['password']
    
    @password.setter
    def password(self, password):
        self.log.debug('Setting password')
        self._config['password'] = password



#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  ListBoxRows for messages.
"""

import logging

import gi
gi.require_versions(
    {
        'Gtk': '3.0',
        'Pango': '1.0'
    }
)
from gi.repository import Gtk, GLib, Pango

from ad1459.formatting import Parser

class MessageRow(Gtk.ListBoxRow):
    """ A ListBoxRow representing a message.

    The attributes for this class are stored directly in the Gtk.Label widgets
    drawn within these rows, as this obviates the need to store extra variables
    in memory.

    We use two copies of the message text, timestamp, and sender. This is to
    display the message differently depending on if this was a server message or
    a normal message. Server messages are combined into one long message, and 
    be collapsed to save space.

    Attributes:
        time (str): The time the message was sent.
        sender (str): The user/network who sent the message. "*" by default.
        text (str): The text of the message.
    """

    def __init__(self):
        Gtk.ListBoxRow.__init__(self)
        self.log = logging.getLogger('ad1459.message-row')

        # We don't want these to be selectable in the list
        self.props.selectable = False
        Gtk.StyleContext.add_class(self.get_style_context(), "message-row")
        self.props.margin = 1
        self.set_margin_start(6)
        self.set_margin_end(6)
        self.set_focus_on_click(False)

        # We use the stack to display server/message rows with a single widget
        self.stack = Gtk.Stack()
        self.add(self.stack)

        # This is the layout for normal messages
        self.message_grid = Gtk.Grid()
        self.message_grid.set_hexpand(True)
        self.message_grid.set_margin_top(1)
        self.message_grid.set_margin_bottom(2)
        self.message_grid.set_margin_start(12)
        self.message_grid.set_margin_end(12)
        self.message_grid.set_column_spacing(12)

        # This is the layout for server messages.
        self.server_grid = Gtk.Grid()
        self.server_grid.set_hexpand(True)
        self.server_grid.set_valign(Gtk.Align.CENTER)
        self.server_grid.set_margin_left(12)
        self.server_grid.set_margin_right(12)
        self.server_grid.set_column_spacing(12)

        # The sender for normal messages
        self.message_sender = Gtk.Label()
        self.message_sender.set_margin_bottom(1)
        self.message_sender.set_selectable(True)
        self.message_sender.set_use_markup(True)
        self.message_sender.set_halign(Gtk.Align.START)
        self.message_sender.set_line_wrap(True)
        self.message_sender.set_line_wrap_mode(Pango.WrapMode.CHAR)
        self.message_sender.props.opacity = 0.8
        self.message_grid.attach(self.message_sender, 0, 0, 1, 1)
        
        # The timestamp for normal messages
        self.message_time = Gtk.Label()
        self.message_time.set_selectable(True)
        self.message_time.set_use_markup(True)
        self.message_time.set_halign(Gtk.Align.END)
        self.message_time.props.opacity = 0.5
        self.message_grid.attach(self.message_time, 1, 0, 1, 1)
        
        # The text for normal messages
        self.message_text = Gtk.Label()
        self.message_text.set_selectable(True)
        self.message_text.set_use_markup(False)
        self.message_text.set_halign(Gtk.Align.START)
        self.message_text.set_line_wrap(True)
        self.message_text.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)
        self.message_text.set_hexpand(True)
        self.message_text.set_margin_start(12)
        self.message_text.props.xalign = 0
        self.message_grid.attach(self.message_text, 0, 1, 2, 1)

        # The sender for server messages
        self.server_sender = Gtk.Label()
        self.server_sender.set_selectable(True)
        self.server_sender.set_use_markup(True)
        self.server_sender.props.xalign = 1
        self.server_sender.props.halign = Gtk.Align.START
        self.server_sender.props.valign = Gtk.Align.START
        self.server_sender.props.opacity = 0.8
        self.server_grid.attach(self.server_sender, 0, 0, 1, 1)

        # The timestamp for server messages
        self.server_time = Gtk.Label()
        self.server_time.set_selectable(True)
        self.server_time.set_use_markup(True)
        self.server_time.set_width_chars(9)
        self.server_time.props.halign = Gtk.Align.END
        self.server_time.props.valign = Gtk.Align.START
        self.server_time.props.opacity = 0.5
        self.server_grid.attach(self.server_time, 2, 0, 1, 1)

        # We need expander + revealer to animate the message contents.
        self.server_message_expander = Gtk.Expander()
        self.server_message_expander.set_expanded(True)
        self.server_message_expander.set_hexpand(True)
        self.server_message_expander.props.opacity = 0.6
        self.server_message_expander.connect('notify::expanded', self.show_hide)
        self.server_grid.attach(self.server_message_expander, 1, 0, 1, 1)
        self.server_revealer = Gtk.Revealer()
        self.server_revealer.set_transition_type(
            Gtk.RevealerTransitionType.SLIDE_DOWN
        )
        self.server_revealer.props.reveal_child = True
        self.server_grid.attach(self.server_revealer, 0, 1, 3, 1)

        # The text for server messages
        self.server_text = Gtk.Label()
        self.server_text.set_selectable(True)
        self.server_text.set_use_markup(False)
        self.server_text.props.halign = Gtk.Align.START
        self.server_text.set_line_wrap(True)
        self.server_text.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)
        self.server_text.set_hexpand(True)
        self.server_text.props.xalign = 0
        self.server_text.props.opacity = 0.7
        self.server_revealer.add(self.server_text)

        # Show everything
        self.message_text.show()
        self.message_time.show()
        self.message_sender.show()
        self.message_grid.show()
        self.server_text.show()
        self.server_time.show()
        self.server_sender.show()
        self.server_grid.show()

        # Keep track of server messages (this is ignored for normal messages)
        self.server_count = 1
        self.server_message_expander.set_label(
            f'{self.server_count} server messages'
        )

        self.parser = Parser()
        self.kind = 'message'

        # Keep track of whether the user toggles the expander
        self.toggled = False
    
    def update_server_message(self, text):
        """ Adds text to a server message.

        We set the updated text (placing it on a newline) and increment the 
        count of the messages. 

        Arguments:
            text (str): The line to add to the server message.
        """
        # We use '/n' because our parsing later on filters out newlines.
        # It will ignore '/n', so we use that and replace it before setting.
        self.text += f'/n{text} '
        self.server_count += 1
        self.server_message_expander.set_label(
            f'{self.server_count} server messages'
        )
        if self.server_count > 10 and not self.toggled:
            self.server_message_expander.set_expanded(False)
    
    # Internal Handlers
    def show_hide(self, expander, data=None):
        """ handler for the expander-revealer above. 
        
        When the expander is toggled, we set the revealer property to the same
        as the expander's property.
        """
        self.server_revealer.props.reveal_child = expander.get_expanded()
        # User toggled manually, remember that.
        self.toggled = True
    
    @property
    def kind(self):
        """str: The type of message this is, based on who sent it and 
        its contents.
        """
        return self._kind
    
    @kind.setter
    def kind(self, kind):
        self._kind = kind
        Gtk.StyleContext.add_class(self.get_style_context(), kind)
        for child in self.stack.get_children():
            self.stack.remove(child)

        if kind == 'server':
            self.stack.add_named(self.server_grid, 'server')
            self.stack.set_visible_child_name('server')
        else:
            self.stack.add_named(self.message_grid, 'message')
            self.stack.set_visible_child_name('message')


    @property
    def time(self):
        """str: The time the message was sent."""
        return self.message_time.get_text()
    
    @time.setter
    def time(self, time):
        self.message_time.set_markup(f'<small><i>{time}</i></small>')
        self.server_time.set_markup(f'<small><i>{time}</i></small>')
    
    @property
    def sender(self):
        """ The sender of the message, or * for none/network. """
        return self.message_sender.get_text()
    
    @sender.setter
    def sender(self, sender):
        self.message_sender.set_markup(f'<small><b>{sender}</b></small>')
        self.server_sender.set_markup(f'<small>{sender}</small>')
    
    @property
    def text(self):
        """ The message text."""
        return self.message_text.get_text()
    
    @text.setter
    def text(self, text):
        """ We need to make some conversions depending on the type of message
        this is.
        """
        self.log.debug('Setting text to %s', text)
        # Action message formatting
        if self.kind == 'action':
            text = f'\x1D{text}\x1D'
        
        # This is our message parsing for formatting and such.
        escaped_text = GLib.markup_escape_text(text, len(text.encode('utf-8')))
        formatted_text = self.parser.parse_text(escaped_text)
        linked_text = self.parser.hyperlinks(formatted_text.replace('/n', ' /n'))

        # Set both of the labels to the same text.
        self.message_text.set_markup(linked_text)
        self.server_text.set_markup(linked_text.replace('/n', '\n'))
    
    def show_all_contents(self):
        self.show_all()



#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  This is the Python package.
"""

from setuptools import setup, find_packages, Command
import subprocess

version = {}
with open('ad1459/__version__.py') as fp:
    exec(fp.read(), version)

with open('README.md') as readme:
    long_description = readme.read()

class Release(Command):
    """ Generate a release and push it to git."""
    description = "Generate a release and push it to git."

    user_options = [
        ('dry-run', None, 'Skip the actual release and do a dry run instead.'),
        ('prerelease', None, 'Release this version as a pre-release.'),
        ('force-version=', None, 'Force the version to update to the given value.')
    ]

    def initialize_options(self):
        self.dry_run = False
        self.prerelease = False
        self.force_version = None
    
    def finalize_options(self):
        if self.force_version:
            if not isinstance(self.force_version, str):
                raise Exception('Please specify the test version to release')

    def run(self):
        command = ['npx', 'standard-version@next']
        if self.dry_run:
            command.append('--dry-run')
        if self.prerelease:
            command.append('--prerelease')
        if self.force_version:
            # See https://github.com/conventional-changelog/standard-version#release-as-a-target-type-imperatively-npm-version-like
            command.append('--release-as')
            command.append(self.force_version)
        print(command)
        subprocess.run(command)
        if not self.dry_run:
            subprocess.run(
                ['git', 'push', '--follow-tags']
            )

setup(
    name='ad1459',
    version=version['__version__'],
    packages=find_packages(),
    scripts=['ad1459/ad1459'],

    # Dependencies
    install_requires=[
      'pydle',
      'pure-sasl',
      'keyring'
    ],

    data_files=[
      ('/usr/share/applications', ['data/in.donotspellitgav.ad1459.desktop']),
      ('/usr/share/icons/hicolor/scalable/apps', [
          'data/in.donotspellitgav.Ad1459.svg',
          'data/in.donotspellitgav.Ad1459.Devel.svg'
      ]),
      ('/usr/share/icons/hicolor/symbolic/apps', [
          'data/in.donotspellitgav.Ad1459-symbolic.svg'
      ])
    ],

    # Commands
    cmdclass={'release': Release},

    # Project Metadata
    author='Gaven Royer',
    author_email='gavroyer@gmail.com',
    description='An IRC Client',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='ad1459 irc client chat gui gtk',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Environment :: X11 Applications :: Gnome",
        "Environment :: X11 Applications :: GTK",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Topic :: Communications :: Chat :: Internet Relay Chat"
    ]
)


#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  This file is the licence for the project.
"""

__license__ = (
'    Mozilla Public License Version 2.0\n'
'==================================\n'
'\n'
'1. Definitions\n'
'--------------\n'
'\n'
'1.1. "Contributor"\n'
'    means each individual or legal entity that creates, contributes to\n'
'    the creation of, or owns Covered Software.\n'
'\n'
'1.2. "Contributor Version"\n'
'    means the combination of the Contributions of others (if any) used\n'
'    by a Contributor and that particular Contributor\'s Contribution.\n'
'\n'
'1.3. "Contribution"\n'
'    means Covered Software of a particular Contributor.\n'
'\n'
'1.4. "Covered Software"\n'
'    means Source Code Form to which the initial Contributor has attached\n'
'    the notice in Exhibit A, the Executable Form of such Source Code\n'
'    Form, and Modifications of such Source Code Form, in each case\n'
'    including portions thereof.\n'
'\n'
'1.5. "Incompatible With Secondary Licenses"\n'
'    means\n'
'\n'
'    (a) that the initial Contributor has attached the notice described\n'
'        in Exhibit B to the Covered Software; or\n'
'\n'
'    (b) that the Covered Software was made available under the terms of\n'
'        version 1.1 or earlier of the License, but not also under the\n'
'        terms of a Secondary License.\n'
'\n'
'1.6. "Executable Form"\n'
'    means any form of the work other than Source Code Form.\n'
'\n'
'1.7. "Larger Work"\n'
'    means a work that combines Covered Software with other material, in\n'
'    a separate file or files, that is not Covered Software.\n'
'\n'
'1.8. "License"\n'
'    means this document.\n'
'\n'
'1.9. "Licensable"\n'
'    means having the right to grant, to the maximum extent possible,\n'
'    whether at the time of the initial grant or subsequently, any and\n'
'    all of the rights conveyed by this License.\n'
'\n'
'1.10. "Modifications"\n'
'    means any of the following:\n'
'\n'
'    (a) any file in Source Code Form that results from an addition to,\n'
'        deletion from, or modification of the contents of Covered\n'
'        Software; or\n'
'\n'
'    (b) any new file in Source Code Form that contains any Covered\n'
'        Software.\n'
'\n'
'1.11. "Patent Claims" of a Contributor\n'
'    means any patent claim(s), including without limitation, method,\n'
'    process, and apparatus claims, in any patent Licensable by such\n'
'    Contributor that would be infringed, but for the grant of the\n'
'    License, by the making, using, selling, offering for sale, having\n'
'    made, import, or transfer of either its Contributions or its\n'
'    Contributor Version.\n'
'\n'
'1.12. "Secondary License"\n'
'    means either the GNU General Public License, Version 2.0, the GNU\n'
'    Lesser General Public License, Version 2.1, the GNU Affero General\n'
'    Public License, Version 3.0, or any later versions of those\n'
'    licenses.\n'
'\n'
'1.13. "Source Code Form"\n'
'    means the form of the work preferred for making modifications.\n'
'\n'
'1.14. "You" (or "Your")\n'
'    means an individual or a legal entity exercising rights under this\n'
'    License. For legal entities, "You" includes any entity that\n'
'    controls, is controlled by, or is under common control with You. For\n'
'    purposes of this definition, "control" means (a) the power, direct\n'
'    or indirect, to cause the direction or management of such entity,\n'
'    whether by contract or otherwise, or (b) ownership of more than\n'
'    fifty percent (50%) of the outstanding shares or beneficial\n'
'    ownership of such entity.\n'
'\n'
'2. License Grants and Conditions\n'
'--------------------------------\n'
'\n'
'2.1. Grants\n'
'\n'
'Each Contributor hereby grants You a world-wide, royalty-free,\n'
'non-exclusive license:\n'
'\n'
'(a) under intellectual property rights (other than patent or trademark)\n'
'    Licensable by such Contributor to use, reproduce, make available,\n'
'    modify, display, perform, distribute, and otherwise exploit its\n'
'    Contributions, either on an unmodified basis, with Modifications, or\n'
'    as part of a Larger Work; and\n'
'\n'
'(b) under Patent Claims of such Contributor to make, use, sell, offer\n'
'    for sale, have made, import, and otherwise transfer either its\n'
'    Contributions or its Contributor Version.\n'
'\n'
'2.2. Effective Date\n'
'\n'
'The licenses granted in Section 2.1 with respect to any Contribution\n'
'become effective for each Contribution on the date the Contributor first\n'
'distributes such Contribution.\n'
'\n'
'2.3. Limitations on Grant Scope\n'
'\n'
'The licenses granted in this Section 2 are the only rights granted under\n'
'this License. No additional rights or licenses will be implied from the\n'
'distribution or licensing of Covered Software under this License.\n'
'Notwithstanding Section 2.1(b) above, no patent license is granted by a\n'
'Contributor:\n'
'\n'
'(a) for any code that a Contributor has removed from Covered Software;\n'
'    or\n'
'\n'
'(b) for infringements caused by: (i) Your and any other third party\'s\n'
'    modifications of Covered Software, or (ii) the combination of its\n'
'    Contributions with other software (except as part of its Contributor\n'
'    Version); or\n'
'\n'
'(c) under Patent Claims infringed by Covered Software in the absence of\n'
'    its Contributions.\n'
'\n'
'This License does not grant any rights in the trademarks, service marks,\n'
'or logos of any Contributor (except as may be necessary to comply with\n'
'the notice requirements in Section 3.4).\n'
'\n'
'2.4. Subsequent Licenses\n'
'\n'
'No Contributor makes additional grants as a result of Your choice to\n'
'distribute the Covered Software under a subsequent version of this\n'
'License (see Section 10.2) or under the terms of a Secondary License (if\n'
'permitted under the terms of Section 3.3).\n'
'\n'
'2.5. Representation\n'
'\n'
'Each Contributor represents that the Contributor believes its\n'
'Contributions are its original creation(s) or it has sufficient rights\n'
'to grant the rights to its Contributions conveyed by this License.\n'
'\n'
'2.6. Fair Use\n'
'\n'
'This License is not intended to limit any rights You have under\n'
'applicable copyright doctrines of fair use, fair dealing, or other\n'
'equivalents.\n'
'\n'
'2.7. Conditions\n'
'\n'
'Sections 3.1, 3.2, 3.3, and 3.4 are conditions of the licenses granted\n'
'in Section 2.1.\n'
'\n'
'3. Responsibilities\n'
'-------------------\n'
'\n'
'3.1. Distribution of Source Form\n'
'\n'
'All distribution of Covered Software in Source Code Form, including any\n'
'Modifications that You create or to which You contribute, must be under\n'
'the terms of this License. You must inform recipients that the Source\n'
'Code Form of the Covered Software is governed by the terms of this\n'
'License, and how they can obtain a copy of this License. You may not\n'
'attempt to alter or restrict the recipients\' rights in the Source Code\n'
'Form.\n'
'\n'
'3.2. Distribution of Executable Form\n'
'\n'
'If You distribute Covered Software in Executable Form then:\n'
'\n'
'(a) such Covered Software must also be made available in Source Code\n'
'    Form, as described in Section 3.1, and You must inform recipients of\n'
'    the Executable Form how they can obtain a copy of such Source Code\n'
'    Form by reasonable means in a timely manner, at a charge no more\n'
'    than the cost of distribution to the recipient; and\n'
'\n'
'(b) You may distribute such Executable Form under the terms of this\n'
'    License, or sublicense it under different terms, provided that the\n'
'    license for the Executable Form does not attempt to limit or alter\n'
'    the recipients\' rights in the Source Code Form under this License.\n'
'\n'
'3.3. Distribution of a Larger Work\n'
'\n'
'You may create and distribute a Larger Work under terms of Your choice,\n'
'provided that You also comply with the requirements of this License for\n'
'the Covered Software. If the Larger Work is a combination of Covered\n'
'Software with a work governed by one or more Secondary Licenses, and the\n'
'Covered Software is not Incompatible With Secondary Licenses, this\n'
'License permits You to additionally distribute such Covered Software\n'
'under the terms of such Secondary License(s), so that the recipient of\n'
'the Larger Work may, at their option, further distribute the Covered\n'
'Software under the terms of either this License or such Secondary\n'
'License(s).\n'
'\n'
'3.4. Notices\n'
'\n'
'You may not remove or alter the substance of any license notices\n'
'(including copyright notices, patent notices, disclaimers of warranty,\n'
'or limitations of liability) contained within the Source Code Form of\n'
'the Covered Software, except that You may alter any license notices to\n'
'the extent required to remedy known factual inaccuracies.\n'
'\n'
'3.5. Application of Additional Terms\n'
'\n'
'You may choose to offer, and to charge a fee for, warranty, support,\n'
'indemnity or liability obligations to one or more recipients of Covered\n'
'Software. However, You may do so only on Your own behalf, and not on\n'
'behalf of any Contributor. You must make it absolutely clear that any\n'
'such warranty, support, indemnity, or liability obligation is offered by\n'
'You alone, and You hereby agree to indemnify every Contributor for any\n'
'liability incurred by such Contributor as a result of warranty, support,\n'
'indemnity or liability terms You offer. You may include additional\n'
'disclaimers of warranty and limitations of liability specific to any\n'
'jurisdiction.\n'
'\n'
'4. Inability to Comply Due to Statute or Regulation\n'
'---------------------------------------------------\n'
'\n'
'If it is impossible for You to comply with any of the terms of this\n'
'License with respect to some or all of the Covered Software due to\n'
'statute, judicial order, or regulation then You must: (a) comply with\n'
'the terms of this License to the maximum extent possible; and (b)\n'
'describe the limitations and the code they affect. Such description must\n'
'be placed in a text file included with all distributions of the Covered\n'
'Software under this License. Except to the extent prohibited by statute\n'
'or regulation, such description must be sufficiently detailed for a\n'
'recipient of ordinary skill to be able to understand it.\n'
'\n'
'5. Termination\n'
'--------------\n'
'\n'
'5.1. The rights granted under this License will terminate automatically\n'
'if You fail to comply with any of its terms. However, if You become\n'
'compliant, then the rights granted under this License from a particular\n'
'Contributor are reinstated (a) provisionally, unless and until such\n'
'Contributor explicitly and finally terminates Your grants, and (b) on an\n'
'ongoing basis, if such Contributor fails to notify You of the\n'
'non-compliance by some reasonable means prior to 60 days after You have\n'
'come back into compliance. Moreover, Your grants from a particular\n'
'Contributor are reinstated on an ongoing basis if such Contributor\n'
'notifies You of the non-compliance by some reasonable means, this is the\n'
'first time You have received notice of non-compliance with this License\n'
'from such Contributor, and You become compliant prior to 30 days after\n'
'Your receipt of the notice.\n'
'\n'
'5.2. If You initiate litigation against any entity by asserting a patent\n'
'infringement claim (excluding declaratory judgment actions,\n'
'counter-claims, and cross-claims) alleging that a Contributor Version\n'
'directly or indirectly infringes any patent, then the rights granted to\n'
'You by any and all Contributors for the Covered Software under Section\n'
'2.1 of this License shall terminate.\n'
'\n'
'5.3. In the event of termination under Sections 5.1 or 5.2 above, all\n'
'end user license agreements (excluding distributors and resellers) which\n'
'have been validly granted by You or Your distributors under this License\n'
'prior to termination shall survive termination.\n'
'\n'
'************************************************************************\n'
'*                                                                      *\n'
'*  6. Disclaimer of Warranty                                           *\n'
'*  -------------------------                                           *\n'
'*                                                                      *\n'
'*  Covered Software is provided under this License on an "as is"       *\n'
'*  basis, without warranty of any kind, either expressed, implied, or  *\n'
'*  statutory, including, without limitation, warranties that the       *\n'
'*  Covered Software is free of defects, merchantable, fit for a        *\n'
'*  particular purpose or non-infringing. The entire risk as to the     *\n'
'*  quality and performance of the Covered Software is with You.        *\n'
'*  Should any Covered Software prove defective in any respect, You     *\n'
'*  (not any Contributor) assume the cost of any necessary servicing,   *\n'
'*  repair, or correction. This disclaimer of warranty constitutes an   *\n'
'*  essential part of this License. No use of any Covered Software is   *\n'
'*  authorized under this License except under this disclaimer.         *\n'
'*                                                                      *\n'
'************************************************************************\n'
'\n'
'************************************************************************\n'
'*                                                                      *\n'
'*  7. Limitation of Liability                                          *\n'
'*  --------------------------                                          *\n'
'*                                                                      *\n'
'*  Under no circumstances and under no legal theory, whether tort      *\n'
'*  (including negligence), contract, or otherwise, shall any           *\n'
'*  Contributor, or anyone who distributes Covered Software as          *\n'
'*  permitted above, be liable to You for any direct, indirect,         *\n'
'*  special, incidental, or consequential damages of any character      *\n'
'*  including, without limitation, damages for lost profits, loss of    *\n'
'*  goodwill, work stoppage, computer failure or malfunction, or any    *\n'
'*  and all other commercial damages or losses, even if such party      *\n'
'*  shall have been informed of the possibility of such damages. This   *\n'
'*  limitation of liability shall not apply to liability for death or   *\n'
'*  personal injury resulting from such party\'s negligence to the       *\n'
'*  extent applicable law prohibits such limitation. Some               *\n'
'*  jurisdictions do not allow the exclusion or limitation of           *\n'
'*  incidental or consequential damages, so this exclusion and          *\n'
'*  limitation may not apply to You.                                    *\n'
'*                                                                      *\n'
'************************************************************************\n'
'\n'
'8. Litigation\n'
'-------------\n'
'\n'
'Any litigation relating to this License may be brought only in the\n'
'courts of a jurisdiction where the defendant maintains its principal\n'
'place of business and such litigation shall be governed by laws of that\n'
'jurisdiction, without reference to its conflict-of-law provisions.\n'
'Nothing in this Section shall prevent a party\'s ability to bring\n'
'cross-claims or counter-claims.\n'
'\n'
'9. Miscellaneous\n'
'----------------\n'
'\n'
'This License represents the complete agreement concerning the subject\n'
'matter hereof. If any provision of this License is held to be\n'
'unenforceable, such provision shall be reformed only to the extent\n'
'necessary to make it enforceable. Any law or regulation which provides\n'
'that the language of a contract shall be construed against the drafter\n'
'shall not be used to construe this License against a Contributor.\n'
'\n'
'10. Versions of the License\n'
'---------------------------\n'
'\n'
'10.1. New Versions\n'
'\n'
'Mozilla Foundation is the license steward. Except as provided in Section\n'
'10.3, no one other than the license steward has the right to modify or\n'
'publish new versions of this License. Each version will be given a\n'
'distinguishing version number.\n'
'\n'
'10.2. Effect of New Versions\n'
'\n'
'You may distribute the Covered Software under the terms of the version\n'
'of the License under which You originally received the Covered Software,\n'
'or under the terms of any subsequent version published by the license\n'
'steward.\n'
'\n'
'10.3. Modified Versions\n'
'\n'
'If you create software not governed by this License, and you want to\n'
'create a new license for such software, you may create and use a\n'
'modified version of this License if you rename the license and remove\n'
'any references to the name of the license steward (except to note that\n'
'such modified license differs from this License).\n'
'\n'
'10.4. Distributing Source Code Form that is Incompatible With Secondary\n'
'Licenses\n'
'\n'
'If You choose to distribute Source Code Form that is Incompatible With\n'
'Secondary Licenses under the terms of this version of the License, the\n'
'notice described in Exhibit B of this License must be attached.\n'
'\n'
'Exhibit A - Source Code Form License Notice\n'
'-------------------------------------------\n'
'\n'
'  This Source Code Form is subject to the terms of the Mozilla Public\n'
'  License, v. 2.0. If a copy of the MPL was not distributed with this\n'
'  file, You can obtain one at http://mozilla.org/MPL/2.0/.\n'
'\n'
'If it is not possible or desirable to put the notice in a particular\n'
'file, then You may include the notice in a location (such as a LICENSE\n'
'file in a relevant directory) where a recipient would be likely to look\n'
'for such a notice.\n'
'\n'
'You may add additional accurate notices of copyright ownership.\n'
'\n'
'Exhibit B - "Incompatible With Secondary Licenses" Notice\n'
'---------------------------------------------------------\n'
'\n'
'  This Source Code Form is "Incompatible With Secondary Licenses", as\n'
'  defined by the Mozilla Public License, v. 2.0.\n'
'\n'
)


#!/usr/bin/env python3

""" AD1459, an IRC Client

  Copyright ©2019-2020 by Gaven Royer

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH REGARD
  TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  SOFTWARE.

  Then entry for using IRC.
"""

import logging

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk

class IrcEntry(Gtk.Entry):
    """ An entry specifically tailored to entering IRC messages. 
    
    Includes tab-complete features.
    """

    
    def __init__(self, window, placeholder='Enter a message'):
        self.log = logging.getLogger('ad1459.ircentry')
        super().__init__()

        self.prematched = False
        self.window = window
        self.set_hexpand(True)
        self.set_placeholder_text(placeholder)
        self.props.show_emoji_icon = True
        self.props.max_width_chars = 5000
        self.possible_completions = []
        self.possible_recents = []
        self.recents_mode = ''

        self.connect('key-press-event', self.on_key_press_event)
        self.connect('notify::text', self.on_text_changed)
    
    def set_text_updown(self, msglist, index):
        """ Sets the text in the entry on up/down.

        Arguments:
            msglist (list): The list of messages to use.
            index (int): The index of the desired message in the list.
        """
        message = msglist[index]
        self.log.debug('Index: %s, Buffer: %s', msglist, index)
        self.set_text(message)
        self.set_position(len(message))
    
    def clamp_index(self, value, largest): 
        """ Clamps the value to between 0 and largest.
        
        Arguments:
            value (int): The value we need clamped.
            largest (int): The highest we accept.
        """
        return sorted((0, value, largest - 1))[1]
    
    def on_text_changed(self, entry, data=None):
        """ Update the text in the channel entry buffer."""
        room = self.window.active_room
        text = self.get_text()
        room.unsent_text = self.get_text()
    
    def on_key_press_event(self, entry, event):
        room = self.window.message_stack.get_visible_child().room

        keys = Gdk.Keymap.get_default()
        mods = keys.get_modifier_state()

        self.log.debug('Keyval: %s', Gdk.keyval_name(event.keyval))
        self.log.debug('Modifiers: %s', mods)
        self.log.debug('Completion debugging:')
        self.log.debug('window recents: %s', self.window.recents)
        self.log.debug('room recents: %s', room.recents)
        self.log.debug('recents_mode: %s', self.recents_mode)
        self.log.debug('possible_completions: %s', self.possible_completions)

        if event.keyval == Gdk.keyval_from_name('Up'):
            if mods == 16:
                if self.recents_mode != 'room':
                    self.recents_mode = 'room'
                    self.possible_recents = room.recents.copy()
                    self.possible_recents.append(self.get_text())
                    self.index = len(self.possible_recents) - 1

                self.index -= 1
                self.index = self.clamp_index(self.index, len(self.possible_recents))
                self.set_text_updown(self.possible_recents, self.index)
                return True
            
            elif mods == 20:
                if self.recents_mode != 'window':
                    self.recents_mode = 'window'
                    self.possible_recents = self.window.recents.copy()
                    self.possible_recents.append(self.get_text())
                    self.index = len(self.possible_recents) - 1

                self.index -= 1
                self.index = self.clamp_index(self.index, len(self.possible_recents))
                self.set_text_updown(self.possible_recents, self.index)
                return True
        
        elif event.keyval == Gdk.keyval_from_name('Down'):
            if mods == 16:
                if self.recents_mode != 'room':
                    self.recents_mode = 'room'
                    self.possible_recents = room.recents.copy()
                    self.possible_recents.append(self.get_text())
                    self.index = len(self.possible_recents) - 1

                self.index += 1
                self.index = self.clamp_index(self.index, len(self.possible_recents))
                self.set_text_updown(self.possible_recents, self.index)
                return True
            
            elif mods == 20:
                if self.recents_mode != 'window':
                    self.recents_mode = 'window'
                    self.possible_recents = self.window.recents.copy()
                    self.possible_recents.append(self.get_text())
                    self.index = len(self.possible_recents) - 1

                self.index += 1
                self.index = self.clamp_index(self.index, len(self.possible_recents))
                self.set_text_updown(self.possible_recents, self.index)
                return True

        # Tab completion
        elif event.keyval == Gdk.keyval_from_name('Tab'):
            # We should currently get the most recent word
            # TODO: Improve this to get the current word at the cursor
            text = self.get_text()
            text_list = text.split()
            
            try:
                current_word = text_list.pop()
            except IndexError:
                return True

            complete_list = room.users.copy()
            for command in room.network.commands:
                complete_list.append(command)
            for user in room.tab_complete:
                if user in complete_list[::-1]:
                    complete_list.remove(user)
                    complete_list.insert(0, user)

            print(room.tab_complete)
            self.log.debug('Completing word %s', current_word)
            if not self.possible_completions:
                self.log.debug('Users to complete from: %s', complete_list)
                for user in complete_list:
                    if user.lower().startswith(current_word.lower()):
                        self.possible_completions.insert(0, user)
            self.log.debug(
                'Possible completions of %s: %s', 
                current_word,
                self.possible_completions
            )
            
            try:
                completion = self.possible_completions.pop()
                self.possible_completions.insert(0, completion)

                if len(text_list) == 0 and not completion.startswith('/'):
                    completion = f'{completion}:'

                completion = f'{completion} '
                text_list.append(completion)
                text = ' '.join(text_list)
                self.set_text(text)
                length = len(text)
                self.set_position(length)
            except IndexError:
                pass
            self.grab_focus_without_selecting()
            return True
        
        self.possible_completions = []
        self.possible_recents = []
        self.recents_mode = ''
        self.index = 0
